///////////////////Software Control Header///////////////////////////////////
// BitStrm.h - Library to bit IO operation.
//
// $LogThisFile$ \Projects\UtilCode\Include\BitStrm.h
// 
// V01.01 02/11/2002 0100
//  File seeking functions added.
// 
// V01.00 05/11/2000 2310
//  Initial revision.
// 
// This code is freely distributed to help the development of complex
// software in any field using C++. The author is not held responsible
// for any damage caused by the usage of the code.
// The author should be acknowledged for the use of this piece of code in
// commercial software.
//
// Bug reports and suggestions are welcomed to report to
//
//          tzzstan@hotmail.com
/////////////////////////////////////////////////////////////////////////////
#ifndef __BIT_STREAMER__
#define __BIT_STREAMER__

#include <fstream.h>
#include <string.h>
#include "TSingLst.h"
#include "TString.h"

// Default buffer block of 16kB
#define BUFFER_LENGTH   16384

/////////////////////////////////////////////////////////////////////////////
// CBufferBlock - Memory block to hold data.
//
// TSTan   05 Nov 2000
/////////////////////////////////////////////////////////////////////////////
class CBufferBlock
{
   char *m_Data;  // Pointer to data
   int m_ByteIndex, m_BitIndex;  // Current bit location

public:
   CBufferBlock();
   CBufferBlock(CBufferBlock &source);
   ~CBufferBlock();

   ////////////////
   // Modifiers
   ////////////////

   int AddBits(int data, int bits);

   CBufferBlock &operator=(CBufferBlock &source);

   ////////////////
   // Accessors
   ////////////////
   
   // Return pointer to memory area.
   char *Data() {return m_Data;}
   int SizeInBytes();
   int SpaceInBits();
};

/////////////////////////////////////////////////////////////////////////////
// CBitStream - IO bit operation engine.
//
// TSTan   05 Nov 2000
/////////////////////////////////////////////////////////////////////////////
class CBitStream
{
protected:
   ////////////////////////////////////////////////////////
   // In bitstream data
   ////////////////////////////////////////////////////////
   char *m_Data;
   int m_SizeInBytes, m_ByteIndex, m_BitIndex;

   ////////////////////////////////////////////////////////
   // Output bitstream data
   ////////////////////////////////////////////////////////
   int m_OByteIndex, m_OBitIndex;

   CBufferBlock *m_ActiveOBuffer;
   CSinglyList<CBufferBlock> m_OutputBuffers;

public:
   CBitStream();
   ~CBitStream();

   ////////////////
   // Modifiers
   ////////////////

   void AddBits(int data, int bits);
   void AddBits(CTString &BitString);
   virtual int GetBits(int bits);
	int GetBigEndian(int bytes);
   void Reset();
   void ResetOutBuffers();
	virtual void Seek(int position);
   void SetInputBuffer(char *buffer, int bytesize);
	void SetBufferIdx(int byteidx, int bitidx);
	void SkipBits(int bits);

   ////////////////
   // Accessors
   ////////////////
   
	void FillArray(char *array, int size);
	virtual int GetOffset() {return 0;}
   int RemainingBits();
	virtual int RdBytesAvailable() {return BUFFER_LENGTH;}

   // Return all output buffers.
   CSinglyList<CBufferBlock> *GetOutputBuffers() {return &m_OutputBuffers;}
};

/////////////////////////////////////////////////////////////////////////////
// CFileBitStream - Bit stream operation on file.
//
// TSTan   09 Sep 2002
/////////////////////////////////////////////////////////////////////////////
class CFileBitStream: public CBitStream
{
	ifstream *m_infile;
	ofstream *m_outfile;
	char *m_inputbuffer;
	streampos m_infilesize, m_totalsize;

public:
	CFileBitStream(char *input = NULL, char *output = NULL);
	~CFileBitStream();

	void Reset();
	void SetInFile(char *filename);
	virtual void Seek(int position);

	virtual int GetBits(int bits);
	virtual int GetOffset();
	virtual int RdBytesAvailable();
};

#endif

