///////////////////Software Control Header///////////////////////////////////
// GIFLib.h - Library to support GIF encoding/decoding engine.
//
// $LogThisFile$ \Projects\UtilCode\Include\GIFLib.h
// 
// V01.01 23/11/2002 1337
// 
// V01.00 17/11/2002 1929
//  Initial revision.
// 
// This code is freely distributed to help the development of complex
// software in any field using C++. The author is not held responsible
// for any damage caused by the usage of the code.
// The author should be acknowledged for the use of this piece of code in
// commercial software.
//
// Bug reports and suggestions are welcomed to report to
//
//          tzzstan@hotmail.com
/////////////////////////////////////////////////////////////////////////////
#ifndef __GIF_LIB__
#define __GIF_LIB__

#include "TString.h"
#include "BitStrm.h"
#include "TSingLst.h"

/////////////////////////////////////////////////////////////////////////////
// CGIF_LZWCode
//
// Desctiption: Object to hold a code word of LZW compressed image.
//
// TSTan   17 Nov 2002
/////////////////////////////////////////////////////////////////////////////
class CGIF_LZWCode
{
public:
	int m_codelength;
	char *m_codes;

	CGIF_LZWCode();
	~CGIF_LZWCode();

	void Reset();
	void AddCode(int code);
	void AddCodes(CGIF_LZWCode &code);
};

/////////////////////////////////////////////////////////////////////////////
// CGIFEngine
//
// Desctiption: Engine to decode/encode GIF bitstream.
//
// TSTan   17 Nov 2002
/////////////////////////////////////////////////////////////////////////////
class CGIFEngine
{
	char *m_image, *m_globaltable;

	int m_lzw_count, m_lzw_filling, m_lzw_oldcode;
	char m_lzw_codelength;
	int m_version, m_width, m_height;
	int m_globalcolor, m_globalcolorsize, m_sort, m_colorres,
		 m_aspectratio, m_backgnd;
	int m_codesize, m_clearcode, m_block_size, m_bitidx, m_gifcode;
	int m_disposal, m_usrinput, m_transcolor, m_transcolorflag,
		 m_delay, m_owneraut;
	char m_owner[9];
	char *m_ownerdata, *m_comment;

	CBitStream *m_iobits;
	CGIF_LZWCode *m_LZWCode, *m_lzw_ptr;
	CSinglyList<char *> m_imglist;

	void ByteInReset();
	char ByteInGet();
	int GetBits(int bits);

	////////////////////////////////////////////////////////
	//                Decoder                             //
	////////////////////////////////////////////////////////
	void DecodeImage();
	void ProcessControl();

	////////////////////////////////////////////////////////
	//                Shared                              //
	////////////////////////////////////////////////////////
	void Reset();

public:

	CGIFEngine();
	~CGIFEngine();

///////////////////
// Data retrieval
///////////////////
	int GetX();
	int GetY();
	char *GetImage();
	int Delay() {return m_delay;}
	int Frames() {return m_imglist.Size();}

///////////////////
// Modifiers
///////////////////
	void ProcessInFile(char *filename);
};

#endif