///////////////////Software Control Header///////////////////////////////////
// TIFFLib.h - Library to support TIFF encoding/decoding engine.
//
// $LogThisFile$ \Projects\UtilCode\Include\TIFFLib.h
// 
// V01.00 03/11/2002 0018
//  Initial revision.
// 
// This code is freely distributed to help the development of complex
// software in any field using C++. The author is not held responsible
// for any damage caused by the usage of the code.
// The author should be acknowledged for the use of this piece of code in
// commercial software.
//
// Bug reports and suggestions are welcomed to report to
//
//          tzzstan@hotmail.com
/////////////////////////////////////////////////////////////////////////////
#ifndef __TIFF_LIB__
#define __TIFF_LIB__

#include "TString.h"
#include "BitStrm.h"
#include "TSingLst.h"

/////////////////////////////////////////////////////////////////////////////
// CTIFF_IFDEntry
//
// Desctiption: Object to hold IFD entry.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
class CTIFF_IFDEntry
{
public:
	int m_tag, m_type;
	long m_count, m_offset;
	char *m_data;

	CTIFF_IFDEntry();
	~CTIFF_IFDEntry();
};

/////////////////////////////////////////////////////////////////////////////
// CTIFF_LZWCode
//
// Desctiption: Object to hold a code word of LZW compressed image.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
class CTIFF_LZWCode
{
public:
	int m_codelength;
	char *m_codes;

	CTIFF_LZWCode();
	~CTIFF_LZWCode();

	void Reset();
	void AddCode(int code);
	void AddCodes(CTIFF_LZWCode &code);
};

/////////////////////////////////////////////////////////////////////////////
// CTIFFEngine
//
// Desctiption: Engine to decode/encode TIFF bitstream.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
class CTIFFEngine
{
	char *m_image, *m_table;

	int m_offset, m_order, m_imagewidth, m_imagelength, m_rowsperstrip,
		m_bitspersample, m_photometricinterpretation, m_samplesperpixel;
	char *m_stripoffsets;
	long m_compression;
	int m_lzw_count, m_lzw_filling, m_lzw_oldcode, m_predictor;
	int m_count;
	char m_control, m_data, m_lzw_codelength;

	CBitStream *m_iobits;
	CTIFF_LZWCode *m_LZWCode, *m_lzw_ptr;
	CSinglyList<CTIFF_IFDEntry> m_ifdentries;

	void ByteInReset();
	char ByteInGet(int arg = 0);
	int GetLength();

	////////////////////////////////////////////////////////
	//                Decoder                             //
	////////////////////////////////////////////////////////
	void DecodeImage();
	void ProcessIFDs();

	////////////////////////////////////////////////////////
	//                Shared                              //
	////////////////////////////////////////////////////////
	int GetEndianWord(char *ptr, int bytes);
	int Endian(int data, int bytes);
	void Reset();

public:

	CTIFFEngine();
	~CTIFFEngine();

///////////////////
// Data retrieval
///////////////////
	int GetX();
	int GetY();
	char *GetImage();

///////////////////
// Modifiers
///////////////////
	void ProcessInFile(char *filename);
};

#endif