///////////////////Software Control Header///////////////////////////////////
// TSingLst.h - Library to support singly list
//
// $LogThisFile$ \Projects\UtilCode\Include\TSingLst.h
// 
// V01.03 01/05/2001 1132
//  m_tail introduced for better tail insertion operation.
// 
// V01.02 19/11/2000 1246
//  Allow pointer node insertion.
//  Operator[] overloaded.
// 
// V01.01 15/10/2000 2337
//  CSinglyList - Remove accept pointer to element(default = m_head).
//                Return true on success.
//                Accept pointer insertion.
// 
// V01.00 10/10/2000 2159
//  Initial revision.
// 
// This code is freely distributed to help the development of complex
// software in any field using C++. The author is not held responsible
// for any damage caused by the usage of the code.
// The author should be acknowledged for the use of this piece of code in
// commercial software.
//
// Bug reports and suggestions are welcomed to report to
//
//          tzzstan@hotmail.com
/////////////////////////////////////////////////////////////////////////////

#ifndef __TS_SList__
#define __TS_SList__

#include <iostream.h>

/////////////////////////////////////////////////////////////////////////////
// CSinglyNode - Node representation of an object type.
//
// Desctiption: Manipulate object in node type.
//
// TSTan   12 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> class CSinglyNode
{
   T *m_data;
   CSinglyNode<T> *m_next;

public:
   CSinglyNode(CSinglyNode<T> &, CSinglyNode<T> *Next = NULL); // copy const
   CSinglyNode(T &, CSinglyNode<T> *Next = NULL);
   CSinglyNode(T *, CSinglyNode<T> *Next = NULL);
   ~CSinglyNode();

///////////////////
// Modifiers
///////////////////

   // Set to next pointer
   void SetNext(CSinglyNode<T> *next) {m_next = next;}
   CSinglyNode<T> &operator=(CSinglyNode<T> &);

///////////////////
// Data retrieval
///////////////////

   // Return data pointer
   T *Data() const {return m_data;}
   // Return next pointer
   CSinglyNode<T> *Next() const {return m_next;}
};

/////////////////////////////////////////////////////////////////////////////
// CSinglyNode(CSinglyNode<T> &original, CSinglyNode<T> *Next)
//
// Desctiption: Copy constructor that creates a copy from original.
//
// original - Original node.
// Next - Optional pointer to next element(Defaulted to NULL).
//
// TSTan   12 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CSinglyNode<T>::CSinglyNode(CSinglyNode<T> &original,
   CSinglyNode<T> *Next)
{
   m_data = new T(*(original.m_data));
   m_next = Next;
}

/////////////////////////////////////////////////////////////////////////////
// operator=(CSinglyNode<T> &original
//
// Desctiption: Assignment with copy of T in node.
//
// Original - Original node.
//
// TSTan   12 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CSinglyNode<T> &CSinglyNode<T>::
   operator=(CSinglyNode<T> &original)
{
   delete m_data;
   m_data = new T(*(original.m_data));
   m_next = NULL;

   return *this;
}

/////////////////////////////////////////////////////////////////////////////
// CSinglyNode(T &Element, CSinglyNode<T> *Next)
//
// Desctiption: Constructor to create copy of T in node.
//
// Element - Reference to the element to be inserted.
// Next - Optional pointer to next element(Defaulted to NULL).
//
// TSTan   12 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CSinglyNode<T>::CSinglyNode(T &Element,
   CSinglyNode<T> *Next)
{
   m_data = new T(Element);
   m_next = Next;
}

/////////////////////////////////////////////////////////////////////////////
// CSinglyNode(T *Element, CSinglyNode<T> *Next)
//
// Desctiption: Constructor to create copy of T in node.
//
// Element - Reference to the element to be inserted.
// Next - Optional pointer to next element(Defaulted to NULL).
//
// TSTan   14 Oct 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CSinglyNode<T>::CSinglyNode(T *Element,
   CSinglyNode<T> *Next)
{
   m_data = Element;
   m_next = Next;
}

/////////////////////////////////////////////////////////////////////////////
// ~CSinglyNode()
//
// Desctiption: Destructor to clean up memory.
//
// TSTan   12 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CSinglyNode<T>::~CSinglyNode()
{
   delete m_data;
}

/////////////////////////////////////////////////////////////////////////////
// CSinglyList - class declaration of CSinglyList.
//
// Desctiption: Maintain objects in singly list.
//
// TSTan   12 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> class CSinglyList
{
   int m_size;
   CSinglyNode<T> *m_head, *m_current, *m_tail;

   CSinglyNode<T> *CopyNodes(CSinglyNode<T> *);

public:
   // HEAD/TAIL to insert element.
   enum Location{
      HEAD,
      TAIL
   };

   CSinglyList(CSinglyList<T> &); // Copy constructor
   CSinglyList();
   ~CSinglyList();

///////////////////
// Modifiers
///////////////////
   T *Insert(T &, enum Location = HEAD);
   T *Insert(T *, enum Location = HEAD);
   T *Insert(T *, int index);
   bool Remove(T *Element = NULL);
   void RemoveAll();
   CSinglyNode<T> *ResetPtr();
   void SetHead(CSinglyNode<T> *NewHead) {m_head = NewHead;}
   CSinglyNode<T> *Next();

   CSinglyList<T> &operator=(CSinglyList<T> &); // Assignment operator

///////////////////
// Data retrieval
///////////////////

   // Return data pointer to current node.
   T *Current() const;
   // Return current pointer.
   CSinglyNode<T> *CurrentPtr() const {return m_current;};
   // Return true if the list is empty.
   bool IsEmpty() const {return !m_head;}
   // Return true if current point to valid element.
   bool IsValid() const {return !(m_current == NULL);}
   // Return the size of list.
   int Size() const {return m_size;}
   T *Exist(T &source);

   T *operator[](int index);
};

/////////////////////////////////////////////////////////////////////////////
// Current()
//
// Desctiption: Return ptr to the first node. NULL if empty
//
// TSTan   25 Jul 2002
/////////////////////////////////////////////////////////////////////////////
template<class T> T *CSinglyList<T>::Current() const
{
	if(m_current)
		return m_current->Data();

	return NULL;
}

/////////////////////////////////////////////////////////////////////////////
// Exist(T &Image)
//
// Desctiption: Return ptr to the found node in list.
//
// TSTan   29 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> T *CSinglyList<T>::Exist(T &Image)
{
   for(ResetPtr(); IsValid(); Next())
      if(*Current() == Image)
         return Current();

   return NULL;
}

/////////////////////////////////////////////////////////////////////////////
// ResetPtr()
//
// Desctiption: Set current pointer to the head of list.
//
// TSTan   13 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CSinglyNode<T> *CSinglyList<T>::ResetPtr()
{
   m_current = m_head;
   return m_current;
}

/////////////////////////////////////////////////////////////////////////////
// Next()
//
// Desctiption: Set current pointer to the next of current node.
//
// TSTan   13 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CSinglyNode<T> *CSinglyList<T>::Next()
{
   if(m_current)
      m_current = m_current->Next();

   return m_current;
}

/////////////////////////////////////////////////////////////////////////////
// CSinglyList(CSinglyList<T> &original)
//
// Desctiption: Copy constructor that creates list from original.
//
// original - Original list.
//
// TSTan   12 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CSinglyList<T>::CSinglyList(CSinglyList<T> &original)
{
   m_head = NULL;
	m_tail = NULL;
   m_current = NULL;
   m_size = 0;
   m_head = CopyNodes(original.m_head);
   ResetPtr();
}

/////////////////////////////////////////////////////////////////////////////
// operator=(CSinglyList<T> &original)
//
// Desctiption: Create list from original.
//
// original - Original list.
//
// TSTan   12 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CSinglyList<T> &CSinglyList<T>::
   operator=(CSinglyList<T> &original)
{
   RemoveAll();
   m_head = CopyNodes(original.m_head);
   ResetPtr();

   return *this;
}

/////////////////////////////////////////////////////////////////////////////
// CopyNodes(CSinglyNode<T> *thisNode)
//
// Desctiption: Function to recursively generates all nodes.
//
// thisNode - Pointer to node to generate
//
// TSTan   12 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CSinglyNode<T> *CSinglyList<T>::
   CopyNodes(CSinglyNode<T> *thisNode)
{
   if(thisNode)
   {
      m_head = CopyNodes(thisNode->Next());
      m_size++;

		CSinglyNode<T> *newNode = new CSinglyNode<T>(*thisNode, m_head);
		if(!newNode->Next())
			m_tail = newNode;
      return newNode;
   }

   return NULL;
}

/////////////////////////////////////////////////////////////////////////////
// CSinglyList()
//
// Desctiption: Constructor to initialize list members.
//
// TSTan   12 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CSinglyList<T>::CSinglyList()
{
   m_head = NULL;
	m_tail = NULL;
   m_size = 0;
}

/////////////////////////////////////////////////////////////////////////////
// ~CSinglyList()
//
// Desctiption: Destructor to clean up memory.
//
// TSTan   12 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CSinglyList<T>::~CSinglyList()
{
   RemoveAll();
}

/////////////////////////////////////////////////////////////////////////////
// RemoveAll()
//
// Desctiption: Empty the list.
//
// TSTan   12 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> void CSinglyList<T>::RemoveAll()
{
   while(m_head)
      Remove();

   m_head = NULL;
	m_tail = NULL;
   m_current = NULL;
   m_size = 0;
}

/////////////////////////////////////////////////////////////////////////////
// Insert(T &Element)
//
// Desctiption: Insert one element into list.
//
// TSTan   12 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> T *CSinglyList<T>::Insert(T &Element, enum Location where)
{
   CSinglyNode<T> *NewNode;

   m_size++;
   if(where == HEAD)
   {
		NewNode = new CSinglyNode<T>(Element, m_head);
      m_head = NewNode;

		if(m_tail == NULL)
			m_tail = m_head;
   }
   else
	{
		NewNode = new CSinglyNode<T>(Element);

		if(m_tail)
		{
			m_tail->SetNext(NewNode);
			m_tail = NewNode;
		}
		else
			m_head = m_tail = NewNode;
	}

   return NewNode->Data();
}

/////////////////////////////////////////////////////////////////////////////
// Insert(T *Element)
//
// Desctiption: Insert one element into list.
//
// TSTan   14 Oct 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> T *CSinglyList<T>::Insert(T *Element, enum Location where)
{
   CSinglyNode<T> *NewNode;

   m_size++;
   if(where == HEAD)
   {
		NewNode = new CSinglyNode<T>(Element, m_head);
      m_head = NewNode;

		if(m_tail == NULL)
			m_tail = m_head;
   }
   else
	{
		NewNode = new CSinglyNode<T>(Element);

		if(m_tail)
		{
			m_tail->SetNext(NewNode);
			m_tail = NewNode;
		}
		else
			m_head = m_tail = NewNode;
	}

   return NewNode->Data();
}

/////////////////////////////////////////////////////////////////////////////
// Insert(T *Element, int index)
//
// Desctiption: Insert one element into list at index location.
//              Insert at the end of list if index out of range.
//
// TSTan   18 Nov 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> T *CSinglyList<T>::Insert(T *Element, int index)
{
   if(m_head && index > 0)
   {
      CSinglyNode<T> *NodePtr = m_head;
      int CurIdx = 1;

      while(CurIdx < index && NodePtr->Next())
      {
         CurIdx++;
         NodePtr = NodePtr->Next();
      }

      CSinglyNode<T> *NewNode = new CSinglyNode<T>(Element, NodePtr->Next());
      NodePtr->SetNext(NewNode);

		if(!NewNode->Next())
			m_tail = NewNode;

		m_size++;
   }
   else
		Insert(Element);

   return Element;
}

/////////////////////////////////////////////////////////////////////////////
// operator[](int index)
//
// Desctiption: Return index element from list. NULL if out of range.
//
// TSTan   15 Nov 2000
//
/////////////////////////////////////////////////////////////////////////////
template<class T> T* CSinglyList<T>::operator[](int index)
{
   CSinglyNode<T> *Head = m_head;

   int CurIndex = 0;
   while(Head && CurIndex != index)
   {
      Head = Head->Next();
      CurIndex++;
   }

   if(Head)
      return Head->Data();
   return NULL;
}

/////////////////////////////////////////////////////////////////////////////
// Remove()
//
// Desctiption: Remove the pointed element from list. Comparing address.
//
// TSTan   12 Aug 2000
//
/////////////////////////////////////////////////////////////////////////////
template<class T> bool CSinglyList<T>::Remove(T *Element)
{
   if(!m_head)
      return false;

   if(Element == NULL)
      Element = m_head->Data();

   CSinglyNode<T> *PrevNode = NULL;
   for(CSinglyNode<T> *CurrentNode = m_head;
      CurrentNode; CurrentNode = CurrentNode->Next())
   {
      if(CurrentNode->Data() != Element)
      {
         PrevNode = CurrentNode;
         continue;
      }

      if(PrevNode)
         PrevNode->SetNext(CurrentNode->Next());
      else
         m_head = CurrentNode->Next();

		if(!CurrentNode->Next())
			m_tail = PrevNode;

      delete CurrentNode;
      m_size--;
      return true;
   }

   return false;
}

#endif

