///////////////////Software Control Header///////////////////////////////////
// TString.h - Library to support char string manipulation
//
// $LogThisFile$ \Projects\UtilCode\Include\TString.h
// 
// V01.06 15/08/2002 2210
//  Added Trim().
// 
// V01.05 29/06/2002 1421
//  Scientific notation display.
//  Functions supporting Replace() added.
// 
// V01.04 21/07/2001 1345
//  TrimRight and TrimLeft added.
// 
// V01.03 14/03/2001 2129
//  ToInteger(int Location) added.
// 
// V01.02 29/10/2000 2308
//  RemoveLeft() and operator<<(float) added.
// 
// V01.01 26/10/2000 2155
//  Various changes to deallocate memory only if exceeding m_Limit.
//  Default to 1024 bytes.
// 
// V01.00 11/10/2000 2154
//  Initial revision.
// 
// This code is freely distributed to help the development of complex
// software in any field using C++. The author is not held responsible
// for any damage caused by the usage of the code.
// The author should be acknowledged for the use of this piece of code in
// commercial software.
//
// Bug reports and suggestions are welcomed to report to
//
//          tzzstan@hotmail.com
/////////////////////////////////////////////////////////////////////////////
#ifndef __TS_STRING__
#define __TS_STRING__

#include <iostream.h>
#include <iomanip.h>
#include <strstrea.h>
#include <ctype.h>
#include <math.h>

/////////////////////////////////////////////////////////////////////////////
// CTString - Class to hold and manipulate character string.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
class CTString
{
   char *m_data;
   int m_Limit, m_Current, m_precision, m_width;
	int m_flag;

public:

	enum str_mode {
		str_scientific = 0x0001,	// Use scientific if exceeding width
		str_always_sci = 0x0002		// Always use scientific
	};
	
	static CTString m_TempStorage;

   CTString(char *string = "", int LimitingSize = 1024);
   CTString(const CTString &);
   ~CTString();

	friend CTString &operator<<(int &Value, CTString &String);

   ////////////////
   // Modifiers
   ////////////////

   CTString &operator=(CTString &);
   CTString &operator=(char *);
	void DeleteLine(int linenumber = 0);
   void InsertAt(int Location, char Character);
   void InsertAt(int Location, CTString &string);
   void InsertAt(int Location, char *string);
   int RemoveAt(int Location);
   void RemoveLeft(int Characters);
	void Removes(int Location, int Total);
	CTString &RemoveWord();
	int Replace(char original, char newchar);
	int Replace(char *oldstr, char *newstr);
	void ResetFlag(int flag);
   void SetAt(int Location, char Character);
	void SetFlag(int flag);
   void SetLimit(int NewLimit) {m_Limit = NewLimit;}
	void SetPrecision(int Precision) {m_precision = Precision;}
	void SetWidth(int Width) {m_width = Width;}
	int ToInteger(int Location = 0);
	int ToHexInteger(int Location = 0);
	double ToDouble(int Location = 0);
   void ToUpper();
	int TrimLeft(int Characters = 0);
	int TrimRight(int Characters = 0);
	int Trim() {return TrimLeft() + TrimRight();}
	CTString &StripLeftChars(int Characters);

   ////////////////
   // Accessors
   ////////////////

   int Size();
   // Return char pointer to string.
   char *Data() {return m_data;}
	int Find(char *string, unsigned int index = 0);
   int GetChar(char Character, int FromLeftRight = 0);
	int Lines();
   bool operator==(CTString &);
   bool operator==(char *);
	bool operator!=(CTString &str) {return !(*this == str);}
	bool operator!=(char *str) {return !(*this == str);}
   bool operator<(CTString &);
   bool operator>(CTString &);
   bool operator>=(CTString &);
   bool operator<=(CTString &);
   CTString &operator+=(char *);
   CTString &operator+=(char);
   CTString &operator+=(CTString &);
   CTString &operator+=(int);
   CTString &operator+=(float);
	CTString &operator+=(double);
   char operator[](int);

   friend ostream& operator<<(ostream &, CTString &);
};

CTString &operator<<(int &Value, CTString &String);

#endif
