///////////////////Software Control Header///////////////////////////////////
// BMPLib.cpp - Library to support BMP encoding/decoding.
//
// $LogThisFile$ \Projects\UtilCode\Lib\BMPLib.cpp
// 
// V01.01 30/11/2002 0209
//  Format image buffer to Windows DIB format.
// 
// V01.00 19/10/2002 1707
//  Initial revision.
// 
// This code is freely distributed to help the development of complex
// software in any field using C++. The author is not held responsible
// for any damage caused by the usage of the code.
// The author should be acknowledged for the use of this piece of code in
// commercial software.
//
// Bug reports and suggestions are welcomed to report to
//
//          tzzstan@hotmail.com
/////////////////////////////////////////////////////////////////////////////
#include "BMPLib.h"

/////////////////////////////////////////////////////////////////////////////
// CBMPEngine()
//
// Description: Contructor to initialize pointers.
//
// TSTan   19 Oct 2002
/////////////////////////////////////////////////////////////////////////////
CBMPEngine::CBMPEngine()
{
	m_image = NULL;
}

/////////////////////////////////////////////////////////////////////////////
// ~CBMPEngine()
//
// Description: Cleanup.
//
// TSTan   19 Oct 2002
/////////////////////////////////////////////////////////////////////////////
CBMPEngine::~CBMPEngine()
{
	Reset();
}

/////////////////////////////////////////////////////////////////////////////
// Decompress4()
//
// Description: Decompress4 RLE encoded 16 color image.
//
// TSTan   19 Oct 2002
/////////////////////////////////////////////////////////////////////////////
void CBMPEngine::Decompress4()
{
	int x, y, idxnow, tabidx;
	int control, getcontrol, nonrepeat, data, curcount;

	getcontrol = 1;
	nonrepeat = 0;
	curcount = 0;
	x = 0;
	int imgW = 3 * m_biWidth;
	if(imgW % 4)
		imgW += 4 - (imgW % 4);

	for(y = m_biHeight - 1; y >= 0; y--)
	{
		idxnow = y * imgW;
		while(1)
		{
			if(getcontrol)
			{
				control = m_iobits->GetBits(8);
				getcontrol = 0;

				if(control == 0)
				{
					int excontrol = m_iobits->GetBits(8);
					if(excontrol == 0)
					{
						getcontrol = 1;
						x = 0;
						break;
					}
					else if(excontrol == 1)
						return;
					else if(excontrol == 2)
					{
						x += m_iobits->GetBits(8) - 1;
						y -= m_iobits->GetBits(8) - 1;
						getcontrol = 1;
						break;
					}
					else
					{
						nonrepeat = excontrol;
						curcount = 0;
					}
				}
				else
				{
					data = m_iobits->GetBits(8);
					curcount = 0;
				}
			}

			if(nonrepeat)
			{
				if(nonrepeat != curcount)
				{
					curcount++;
					tabidx = m_iobits->GetBits(4);
				}

				if(nonrepeat == curcount)
				{
					if(nonrepeat % 4)
					{
						int extrabyte = 4 - nonrepeat % 4;
						m_iobits->GetBits(extrabyte << 2);
					}
					nonrepeat = 0;
					getcontrol = 1;
				}
			}
			else	// control repeat
			{
				if(control != curcount)
				{
					tabidx = data >> ((curcount & 0x01) << 2);
					tabidx &= 0x0f;
					curcount++;
				}

				if(control == curcount)
					getcontrol = 1;
			}

			tabidx *= 3;
			m_image[idxnow++] = m_table[tabidx + 2];
			m_image[idxnow++] = m_table[tabidx + 1];
			m_image[idxnow++] = m_table[tabidx];
			x++;
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
// Decompress8()
//
// Description: Decompress RLE encoded 256 color image.
//
// TSTan   19 Oct 2002
/////////////////////////////////////////////////////////////////////////////
void CBMPEngine::Decompress8()
{
	int x, y, idxnow, tabidx;
	int control, getcontrol, nonrepeat, data, curcount;

	getcontrol = 1;
	nonrepeat = 0;
	curcount = 0;
	x = 0;
	int imgW = 3 * m_biWidth;
	if(imgW % 4)
		imgW += 4 - (imgW % 4);

	for(y = m_biHeight - 1; y >= 0; y--)
	{
		idxnow = y * imgW;
		while(1)
		{
			if(getcontrol)
			{
				control = m_iobits->GetBits(8);
				getcontrol = 0;

				if(control == 0)
				{
					int excontrol = m_iobits->GetBits(8);
					if(excontrol == 0)
					{
						getcontrol = 1;
						x = 0;
						break;
					}
					else if(excontrol == 1)
						return;
					else if(excontrol == 2)
					{
						x += m_iobits->GetBits(8) - 1;
						y -= m_iobits->GetBits(8) - 1;
						getcontrol = 1;
						break;
					}
					else
					{
						nonrepeat = excontrol;
						curcount = 0;
					}
				}
				else
				{
					data = m_iobits->GetBits(8);
					curcount = 0;
				}
			}

			if(nonrepeat)
			{
				if(nonrepeat != curcount)
				{
					curcount++;
					tabidx = m_iobits->GetBits(8);
				}

				if(nonrepeat == curcount)
				{
					if(nonrepeat % 2)
						m_iobits->GetBits(8);

					nonrepeat = 0;
					getcontrol = 1;
				}
			}
			else	// control repeat
			{
				if(control != curcount)
				{
					tabidx = data;
					curcount++;
				}

				if(control == curcount)
					getcontrol = 1;
			}

			tabidx *= 3;
			m_image[idxnow++] = m_table[tabidx + 2];
			m_image[idxnow++] = m_table[tabidx + 1];
			m_image[idxnow++] = m_table[tabidx];
			x++;
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
// GetImage()
//
// Description: Return pointer to image RGB array.
//
// TSTan   19 Oct 2002
/////////////////////////////////////////////////////////////////////////////
char *CBMPEngine::GetImage()
{
	return m_image;
}

/////////////////////////////////////////////////////////////////////////////
// GetX()
//
// Description: Return width of image.
//
// TSTan   19 Oct 2002
/////////////////////////////////////////////////////////////////////////////
int CBMPEngine::GetX()
{
	return m_biWidth;
}

/////////////////////////////////////////////////////////////////////////////
// GetY()
//
// Description: Return height of image.
//
// TSTan   19 Oct 2002
/////////////////////////////////////////////////////////////////////////////
int CBMPEngine::GetY()
{
	return m_biHeight;
}

/////////////////////////////////////////////////////////////////////////////
// LittleEnd(int data, int bytes)
//
// Description: Reverse data bytes to form Little Endian word.
//
// data - data in Big Endian.
// bytes - length of data in bytes.
//
// TSTan   19 Oct 2002
/////////////////////////////////////////////////////////////////////////////
int CBMPEngine::LittleEnd(int data, int bytes)
{
	int retdata = 0;
	while(bytes--)
	{
		retdata <<= 8;
		retdata |= data & 0x00ff;
		data >>= 8;
	}

	return retdata;
}

/////////////////////////////////////////////////////////////////////////////
// LoadUncompress()
//
// Description: Decode the uncompress BMP file.
//
// TSTan   19 Oct 2002
/////////////////////////////////////////////////////////////////////////////
void CBMPEngine::LoadUncompress()
{
	int x, y, idxnow, red, green, blue, tabidx, padbits;

	idxnow = m_biWidth * m_biBitCount;
	padbits = idxnow % 32;
	if(padbits)
		padbits = 32 - padbits;

	for(y = m_biHeight - 1; y >= 0; y--)
	{
		idxnow = 3 * y * m_biWidth;
		for(x = 0; x < m_biWidth; x++)
		{
			switch(m_biBitCount)
			{
			case 24:
				blue = m_iobits->GetBits(8);
				green = m_iobits->GetBits(8);
				red = m_iobits->GetBits(8);
				break;
			case 8:
				tabidx = m_iobits->GetBits(8) * 3;
				red = m_table[tabidx];
				green = m_table[tabidx + 1];
				blue = m_table[tabidx + 2];
				break;
			case 4:
				tabidx = m_iobits->GetBits(4) * 3;
				red = m_table[tabidx];
				green = m_table[tabidx + 1];
				blue = m_table[tabidx + 2];
				break;
			case 1:
				tabidx = m_iobits->GetBits(1) * 3;
				red = m_table[tabidx];
				green = m_table[tabidx + 1];
				blue = m_table[tabidx + 2];
				break;
			}

			m_image[idxnow++] = blue;
			m_image[idxnow++] = green;
			m_image[idxnow++] = red;
		}
		m_iobits->GetBits(padbits);
	}
}

/////////////////////////////////////////////////////////////////////////////
// ProcessInFile(char *filename)
//
// Description: Decode the BMP file.
//
// filename - BMP file name.
//
// TSTan   19 Oct 2002
/////////////////////////////////////////////////////////////////////////////
void CBMPEngine::ProcessInFile(char *filename)
{
	Reset();

	CFileBitStream iobitstream(filename);
	m_iobits = &iobitstream;

	if(iobitstream.RdBytesAvailable())
	{
		// Get header
		if(m_iobits->GetBits(16) != 0x424d)	// "BM"
			return;
		m_bfSize = LittleEnd(m_iobits->GetBits(32), 4);
		if(m_iobits->GetBits(32) != 0)
			return;
		m_bfOffBits = LittleEnd(m_iobits->GetBits(32), 4);

		/*typedef struct tagBITMAPINFOHEADER {
				DWORD   biSize;
				LONG    biWidth;
				LONG    biHeight;
				WORD    biPlanes;
				WORD    biBitCount;
				DWORD   biCompression;
				DWORD   biSizeImage;
				LONG    biXPelsPerMeter;
				LONG    biYPelsPerMeter;
				DWORD   biClrUsed;
				DWORD   biClrImportant;
		} BITMAPINFOHEADER;*/

		// BM info
		m_biSize = LittleEnd(m_iobits->GetBits(32), 4);
		m_biWidth = LittleEnd(m_iobits->GetBits(32), 4);
		m_biHeight = LittleEnd(m_iobits->GetBits(32), 4);
		m_biPlanes = LittleEnd(m_iobits->GetBits(16), 2);
		// 1, 4, 8, 24 - bit for each pixel
		m_biBitCount = LittleEnd(m_iobits->GetBits(16), 2);
		// Compression method
		// BI_RLE8, BI_RLE4, BI_RGB
		m_biCompression = LittleEnd(m_iobits->GetBits(32), 4);
		m_biSizeImage = LittleEnd(m_iobits->GetBits(32), 4);
		m_biXPelsPerMeter = LittleEnd(m_iobits->GetBits(32), 4);
		m_biYPelsPerMeter = LittleEnd(m_iobits->GetBits(32), 4);
		// Number of colors in table - 0 = all
		m_biClrUsed = LittleEnd(m_iobits->GetBits(32), 4);
		// Number of important colors, 0 = all
		m_biClrImportant = LittleEnd(m_iobits->GetBits(32), 4);

		// Read color table
		m_table = NULL;
		if(m_biBitCount != 24)
		{
			m_table = new char[3 * (1 << m_biBitCount)];
			int idx;
			int colors = m_biClrUsed;
			if(colors == 0)
				colors = 1 << m_biBitCount;
			for(idx = 0; idx < colors; idx++)
			{
				m_table[3 * idx + 2] = m_iobits->GetBits(8);
				m_table[3 * idx + 1] = m_iobits->GetBits(8);
				m_table[3 * idx] = m_iobits->GetBits(8);
				m_iobits->GetBits(8);
			}
		}

		int imgW = 3 * m_biWidth;
		if(imgW % 4)
			imgW += 4 - (imgW % 4);

		m_image = new char[imgW * m_biHeight];
		if(m_biCompression == bi_rgb)
			LoadUncompress();
		else if(m_biCompression == bi_rle4)
			Decompress4();
		else if(m_biCompression == bi_rle8)
			Decompress8();

		if(m_table)
			delete m_table;
	}
}

/////////////////////////////////////////////////////////////////////////////
// Reset()
//
// Description: Reset to prepare for new image.
//
// TSTan   19 Oct 2002
/////////////////////////////////////////////////////////////////////////////
void CBMPEngine::Reset()
{
	if(m_image)
		delete m_image;
	m_image = NULL;
}
