///////////////////Software Control Header///////////////////////////////////
// GIFLib.cpp - Library to support GIF encoding/decoding.
//
// $LogThisFile$ \Projects\UtilCode\Lib\GIFLib.cpp
// 
// V01.01 30/11/2002 0157
//  Format image buffer to Windows DIB format.
// 
// V01.00 17/11/2002 1929
//  Initial revision.
// 
// This code is freely distributed to help the development of complex
// software in any field using C++. The author is not held responsible
// for any damage caused by the usage of the code.
// The author should be acknowledged for the use of this piece of code in
// commercial software.
//
// Bug reports and suggestions are welcomed to report to
//
//          tzzstan@hotmail.com
/////////////////////////////////////////////////////////////////////////////
#include "GIFLib.h"

/////////////////////////////////////////////////////////////////////////////
// CGIF_LZWCode()
//
// Description: LZW code word object constructor.
//
// TSTan   17 Nov 2002
/////////////////////////////////////////////////////////////////////////////
CGIF_LZWCode::CGIF_LZWCode()
{
	m_codes = NULL;
}

/////////////////////////////////////////////////////////////////////////////
// ~CGIF_LZWCode()
//
// Description: Cleanup.
//
// TSTan   17 Nov 2002
/////////////////////////////////////////////////////////////////////////////
CGIF_LZWCode::~CGIF_LZWCode()
{
	Reset();
}

/////////////////////////////////////////////////////////////////////////////
// AddCode(int code)
//
// Description: Add a code into code word.
//
// code - code to add.
//
// TSTan   17 Nov 2002
/////////////////////////////////////////////////////////////////////////////
void CGIF_LZWCode::AddCode(int code)
{
	if(m_codes == NULL)
	{
		m_codes = new char;
		m_codelength = 1;
		m_codes[0] = code;
	}
	else
	{
		char *tmpcodes = new char[m_codelength + 1];
		int idx;
		for(idx = 0; idx < m_codelength; idx++)
			tmpcodes[idx] = m_codes[idx];
		tmpcodes[m_codelength] = code;
		m_codelength++;

		delete []m_codes;
		m_codes = tmpcodes;
	}
}

/////////////////////////////////////////////////////////////////////////////
// AddCodes(CGIF_LZWCode &code)
//
// Description: Append another object's code word.
//
// code - the source object.
//
// TSTan   17 Nov 2002
/////////////////////////////////////////////////////////////////////////////
void CGIF_LZWCode::AddCodes(CGIF_LZWCode &code)
{
	int idx;
	char *codeptr;

	if(m_codes == NULL)
	{
		m_codes = new char[code.m_codelength];
		m_codelength = code.m_codelength;
		codeptr = m_codes;
	}
	else
	{
		char *tmpcodes = new char[m_codelength + code.m_codelength];
		for(idx = 0; idx < m_codelength; idx++)
			tmpcodes[idx] = m_codes[idx];

		delete []m_codes;
		m_codes = tmpcodes;
		codeptr = &m_codes[m_codelength];
		m_codelength += code.m_codelength;
	}

	for(idx = 0; idx < code.m_codelength; idx++)
	{
		*codeptr = code.m_codes[idx];
		codeptr++;
	}
}

/////////////////////////////////////////////////////////////////////////////
// Reset()
//
// Description: Reset to prepare for new code word.
//
// TSTan   17 Nov 2002
/////////////////////////////////////////////////////////////////////////////
void CGIF_LZWCode::Reset()
{
	if(m_codes)
		delete []m_codes;
	m_codes = NULL;
}

/////////////////////////////////////////////////////////////////////////////
// CGIFEngine()
//
// Description: Contructor to initialize pointers.
//
// TSTan   17 Nov 2002
/////////////////////////////////////////////////////////////////////////////
CGIFEngine::CGIFEngine()
{
	m_globaltable = NULL;
	m_image = NULL;
	m_LZWCode = NULL;
	m_ownerdata = NULL;
	m_comment = NULL;
}

/////////////////////////////////////////////////////////////////////////////
// ~CGIFEngine()
//
// Description: Cleanup.
//
// TSTan   17 Nov 2002
/////////////////////////////////////////////////////////////////////////////
CGIFEngine::~CGIFEngine()
{
	Reset();
}

/////////////////////////////////////////////////////////////////////////////
// ByteInGet()
//
// Description: Return a byte from image with LZW decompressed.
//
// TSTan   17 Nov 2002
/////////////////////////////////////////////////////////////////////////////
char CGIFEngine::ByteInGet()
{
	char retvalue;
	int code, idx;

	if(!m_lzw_count)	// Get new codes
	{
		code = GetBits(m_lzw_codelength);
		if(code == m_clearcode)	// ClearCode
		{
			// Initialize
			for(idx = 0; idx < 4096; idx++)
			{
				m_LZWCode[idx].Reset();
				if(idx < m_clearcode)
					m_LZWCode[idx].AddCode(idx);
				m_lzw_codelength = m_codesize;
			}
			code = GetBits(m_lzw_codelength);
			if(code == m_clearcode + 1)	// EOI
				return 0;
			m_lzw_ptr = &m_LZWCode[code];
			m_lzw_count = m_lzw_ptr->m_codelength;
			m_lzw_oldcode = code;
			m_lzw_filling = m_clearcode + 2;
		}
		else
		{
			if(m_LZWCode[code].m_codes)	// In table
			{
				m_lzw_ptr = &m_LZWCode[code];
				m_LZWCode[m_lzw_filling].AddCodes(m_LZWCode[m_lzw_oldcode]);
				m_LZWCode[m_lzw_filling].AddCode(m_lzw_ptr->m_codes[0]);
			}
			else
			{
				// Add string to table
				m_lzw_ptr = &m_LZWCode[m_lzw_filling];
				m_lzw_ptr->AddCodes(m_LZWCode[m_lzw_oldcode]);
				m_lzw_ptr->AddCode(m_LZWCode[m_lzw_oldcode].m_codes[0]);
			}
			m_lzw_count = m_lzw_ptr->m_codelength;
			m_lzw_oldcode = code;

			switch(m_lzw_filling)
			{
			case 7:
			case 15:
			case 31:
			case 63:
			case 127:
			case 255:
			case 511:
			case 1023:
			case 2047:
				m_lzw_codelength++;
			}
			m_lzw_filling++;
		}
	}

	retvalue = m_lzw_ptr->m_codes[m_lzw_ptr->m_codelength - m_lzw_count];
	m_lzw_count--;

	return retvalue;
}

/////////////////////////////////////////////////////////////////////////////
// ByteInReset()
//
// Description: Prepare for new image block.
//
// TSTan   17 Nov 2002
/////////////////////////////////////////////////////////////////////////////
void CGIFEngine::ByteInReset()
{
	m_clearcode = 1 << (m_codesize - 1);
	m_lzw_count = 0;
	m_block_size = 0;
	m_lzw_codelength = m_codesize;
	m_bitidx = 0;
}

/////////////////////////////////////////////////////////////////////////////
// DecodeImage()
//
// Description: Decode the image into RGB array.
//
// TSTan   17 Nov 2002
/////////////////////////////////////////////////////////////////////////////
void CGIFEngine::DecodeImage()
{
	int startx, starty, width, height, localtable, interlace, sort, tablesize,
		xbyteswpad;

	xbyteswpad = 3 * m_width;
	switch(xbyteswpad % 4)
	{
	case 1:
		xbyteswpad += 3;
		break;
	case 2:
		xbyteswpad += 2;
		break;
	case 3:
		xbyteswpad++;
		break;
	}

	if(!m_image)
		m_image = new char[xbyteswpad * m_height];
	else
	{
		// Copy the previous frame;
		char *newimg = new char[xbyteswpad * m_height];
		char *srcptr = m_image;
		char *destptr = newimg;
		for(starty = 0; starty < m_height; starty++)
		{
			for(startx = 0; startx < xbyteswpad; startx++)
			{
				*destptr = *srcptr;
				destptr++;
				srcptr++;
			}
		}
		m_image = newimg;
	}

	startx = m_iobits->GetBigEndian(2);
	starty = m_iobits->GetBigEndian(2);
	width = m_iobits->GetBigEndian(2);
	height = m_iobits->GetBigEndian(2);
	localtable = m_iobits->GetBits(1);
	interlace = m_iobits->GetBits(1);
	sort = m_iobits->GetBits(1);
	m_iobits->GetBits(2);
	tablesize = m_iobits->GetBits(3);

	char *localcolor = NULL;
	char *usetable = m_globaltable;
	if(localtable)
	{
		int entries = 1 << (tablesize + 1);
		localcolor = new char[3 * entries];
		int idx;
		for(idx = 0; idx < entries; idx++)
		{
			localcolor[3 * idx] = m_iobits->GetBits(8);
			localcolor[3 * idx + 1] = m_iobits->GetBits(8);
			localcolor[3 * idx + 2] = m_iobits->GetBits(8);
		}

		usetable = localcolor;
	}

	m_codesize = m_iobits->GetBits(8) + 1;
	ByteInReset();

	int x, y, color, ystep, laceloop;
	char *imgptr;
	m_LZWCode = new CGIF_LZWCode[4096];

	int interlacelooping = interlace? 4: 1;

	for(laceloop = 0; laceloop < interlacelooping; laceloop++)
	{
		ystep = 1;
		y = 0;
		if(interlace)
		{
			switch(laceloop)
			{
			case 0:
				ystep = 8;
				break;
			case 1:
				y = 4;
				ystep = 8;
				break;
			case 2:
				y = 2;
				ystep = 4;
				break;
			case 3:
				y = 1;
				ystep = 2;
				break;
			}
		}
		for(; y < height; y += ystep)
		{
			//imgptr = &m_image[3 * ((y + starty) * m_width + startx)];
			imgptr = &m_image[xbyteswpad * (y + starty) + 3 * startx];
			for(x = 0; x < width; x++)
			{
				color = ByteInGet() & 0xff;
				if((color == m_transcolor) && m_transcolorflag)
				{
					imgptr++;
					imgptr++;
					imgptr++;
				}
				else
				{
					*imgptr = usetable[3 * color + 2] ;
					imgptr++;
					*imgptr = usetable[3 * color + 1];
					imgptr++;
					*imgptr = usetable[3 * color];
					imgptr++;
				}
			}
		}
	}

	if(localcolor)
		delete localcolor;

	delete []m_LZWCode;

	while(m_block_size || m_bitidx)
		GetBits(1);

	m_iobits->GetBits(8);	// 0 - Terminating block
	m_imglist.Insert(m_image, CSinglyList<char *>::TAIL);
}

/////////////////////////////////////////////////////////////////////////////
// GetBits(int bits)
//
// Description: Return rearranged bit stream from file.
//              GIF arranges bit stream in file from LSB of byte to MSB.
//              Opposite direction compared to TIFF and JPG.
//
// bit - number of bits.
//
// TSTan   17 Nov 2002
/////////////////////////////////////////////////////////////////////////////
int CGIFEngine::GetBits(int bits)
{
	int data = 0;
	int realbits = bits;

	while(bits--)
	{
		if(!m_bitidx)
		{
			if(!m_block_size)
				m_block_size = m_iobits->GetBits(8);

			m_gifcode = m_iobits->GetBits(8);
			m_bitidx = 8;
			m_block_size--;
		}
		data <<= 1;
		if(m_gifcode & (1 << (8 - m_bitidx)))
			data |= 1;
		m_bitidx--;
	}

	int retvalue = 0;
	for(bits = 0; bits < realbits; bits++)
	{
		retvalue <<= 1;
		retvalue |= data & 1;
		data >>= 1;
	}

	return retvalue;
}

/////////////////////////////////////////////////////////////////////////////
// GetImage()
//
// Description: Return pointer to image RGB array.
//
// TSTan   17 Nov 2002
/////////////////////////////////////////////////////////////////////////////
char *CGIFEngine::GetImage()
{
	m_imglist.Next();
	if(!m_imglist.IsValid())
		m_imglist.ResetPtr();

	if(m_imglist.IsValid())
		m_image = *(m_imglist.Current());

	return m_image;
}

/////////////////////////////////////////////////////////////////////////////
// GetX()
//
// Description: Return width of image.
//
// TSTan   17 Nov 2002
/////////////////////////////////////////////////////////////////////////////
int CGIFEngine::GetX()
{
	return m_width;
}

/////////////////////////////////////////////////////////////////////////////
// GetY()
//
// Description: Return height of image.
//
// TSTan   17 Nov 2002
/////////////////////////////////////////////////////////////////////////////
int CGIFEngine::GetY()
{
	return m_height;
}

/////////////////////////////////////////////////////////////////////////////
// ProcessInFile(char *filename)
//
// Description: Decode the GIF file.
//
// filename - GIF file name.
//
// TSTan   17 Nov 2002
/////////////////////////////////////////////////////////////////////////////
void CGIFEngine::ProcessInFile(char *filename)
{
	Reset();

	CFileBitStream iobitstream(filename);
	m_iobits = &iobitstream;

	if(iobitstream.RdBytesAvailable())
	{
		// Get header
		if(m_iobits->GetBits(24) != 0x474946)	// GIF
			return;
		m_version = m_iobits->GetBits(24);
		if(m_version != 0x383761 && m_version != 0x383961)	// 87a, 89a
			return;

		m_width = m_iobits->GetBigEndian(2);
		m_height = m_iobits->GetBigEndian(2);

		m_globalcolor = m_iobits->GetBits(1);
		m_colorres = m_iobits->GetBits(3);
		m_sort = m_iobits->GetBits(1);
		m_globalcolorsize = m_iobits->GetBits(3);
		m_backgnd = m_iobits->GetBits(8);
		m_aspectratio = m_iobits->GetBits(8);

		if(m_globalcolor)
		{
			int entries = 1 << (m_globalcolorsize + 1);
			m_globaltable = new char[3 * entries];
			int idx;
			for(idx = 0; idx < entries; idx++)
			{
				m_globaltable[3 * idx] = m_iobits->GetBits(8);
				m_globaltable[3 * idx + 1] = m_iobits->GetBits(8);
				m_globaltable[3 * idx + 2] = m_iobits->GetBits(8);
			}
		}

		int exitloop = 0;
		m_transcolorflag = 0;
		do
		{
			// Reseting frame dependent variables
			int type = m_iobits->GetBits(8);
			switch(type)
			{
			case 0x2c:	// Image separator
				DecodeImage();
				break;
			case 0x21:
				ProcessControl();
				break;
			default:
				exitloop = 1;
			}
		}while(!exitloop);
	}
}

void CGIFEngine::ProcessControl()
{
	int subblk = m_iobits->GetBits(8);
	if(subblk == 0xf9)
	{
		if(m_iobits->GetBits(8) != 0x04)
			return;

		m_iobits->GetBits(3);
		m_disposal = m_iobits->GetBits(3);
		m_usrinput = m_iobits->GetBits(1);
		m_transcolorflag = m_iobits->GetBits(1);
		m_delay = m_iobits->GetBigEndian(2);
		m_transcolor = m_iobits->GetBits(8);

		m_iobits->GetBits(8);
	}
	else if(subblk == 0xfe)	// Comment
	{
		int size = m_iobits->GetBits(8) + 1;	// with terminating NULL
		if(m_comment)
			delete []m_comment;
		m_comment = new char[size];
		m_iobits->FillArray(m_comment, size);
	}
	else if(subblk == 0xff)
	{
		if(m_iobits->GetBits(8) != 0x0b)
			return;

		m_iobits->FillArray(m_owner, 8);
		m_owner[8] = 0;
		m_owneraut = m_iobits->GetBits(24);

		int privatelength = m_iobits->GetBits(8);
		if(privatelength)
		{
			if(m_ownerdata)
				delete []m_ownerdata;
			m_ownerdata = new char[privatelength];
			m_iobits->FillArray(m_ownerdata, privatelength);
		}
		m_iobits->GetBits(8);
	}
}

/////////////////////////////////////////////////////////////////////////////
// Reset()
//
// Description: Reset to prepare for new image.
//
// TSTan   17 Nov 2002
/////////////////////////////////////////////////////////////////////////////
void CGIFEngine::Reset()
{
	if(m_globaltable)
		delete []m_globaltable;
	m_globaltable = NULL;

	if(m_ownerdata)
		delete []m_ownerdata;
	m_ownerdata = NULL;

	if(m_comment)
		delete []m_comment;
	m_comment = NULL;

	for(m_imglist.ResetPtr(); m_imglist.IsValid(); m_imglist.Next())
	{
		m_image = *(m_imglist.Current());
		delete []m_image;
	}
	m_imglist.RemoveAll();

	m_image = NULL;
}
