///////////////////Software Control Header///////////////////////////////////
// PCXLib.cpp - Library to support PCX encoding/decoding.
//
// $LogThisFile$ \Projects\UtilCode\Lib\PCXLib.cpp
// 
// V01.00 30/11/2002 0232
//  Format image buffer to Windows DIB format.
// 
// This code is freely distributed to help the development of complex
// software in any field using C++. The author is not held responsible
// for any damage caused by the usage of the code.
// The author should be acknowledged for the use of this piece of code in
// commercial software.
//
// Bug reports and suggestions are welcomed to report to
//
//          tzzstan@hotmail.com
/////////////////////////////////////////////////////////////////////////////
#include "PCXLib.h"

/////////////////////////////////////////////////////////////////////////////
// CPCXEngine()
//
// Description: Contructor to initialize pointers.
//
// TSTan   20 Oct 2002
/////////////////////////////////////////////////////////////////////////////
CPCXEngine::CPCXEngine()
{
	m_table = NULL;
	m_image = NULL;
}

/////////////////////////////////////////////////////////////////////////////
// ~CPCXEngine()
//
// Description: Cleanup.
//
// TSTan   20 Oct 2002
/////////////////////////////////////////////////////////////////////////////
CPCXEngine::~CPCXEngine()
{
	Reset();
}

/////////////////////////////////////////////////////////////////////////////
// DecodeLine()
//
// Description: Decompress a line.
//
// TSTan   20 Oct 2002
/////////////////////////////////////////////////////////////////////////////
void CPCXEngine::DecodeLine(char *buffer)
{
	int x;
	int control, data;

	int bytes = m_bytesperline * m_nplanes;

	control = 0;
	for(x = 0; x < bytes; x++)
	{
		if(!control)
		{
			control = m_iobits->GetBits(8);
			if((control & 0xc0) != 0xc0)
			{
				data = control;
				control = 1;
			}
			else
			{
				data = m_iobits->GetBits(8);
				control &= 0x3f;
			}
		}
		
		control--;

		*buffer = data;
		buffer++;
	}
}

/////////////////////////////////////////////////////////////////////////////
// Decompress()
//
// Description: Decompress RLE image.
//
// TSTan   20 Oct 2002
/////////////////////////////////////////////////////////////////////////////
void CPCXEngine::Decompress(char *imgbuf)
{
	int x, y, idxnow, tabidx;
	int idx, offset;

	int width = GetX();
	int height = GetY();
	int imgW = width;
	if(imgW % 4)
		imgW += 4 - (imgW % 4);
	int imgW3 = 3 * width;
	if(imgW3 % 4)
		imgW3 += 4 - (imgW3 % 4);

	char *buffer = new char [m_bytesperline * m_nplanes];

	for(y = 0; y < height; y++)
	{
		if(m_bitsperpixel == 8 && m_nplanes == 1)	// 8bit
			idxnow = y * imgW;
		else
			idxnow = y * imgW3;

		DecodeLine(buffer);
		for(x = 0; x < width; x++)
		{
			tabidx = 0;
			for(idx = 0; idx < m_nplanes; idx++)
			{
				tabidx <<= m_bitsperpixel;
				offset = m_bytesperline * (m_nplanes - idx - 1);

				switch(m_bitsperpixel)
				{
				case 1:
					tabidx |= (buffer[x / 8 + offset] >> (7 - x % 8)) & 1;
					break;
				case 2:
					tabidx |= (buffer[x / 4 + offset] >> ((3 - x % 4) * 2)) & 0x3;
					break;
				case 4:
					tabidx |= (buffer[x / 2 + offset] >> ((1 - x % 2) * 4)) & 0xf;
					break;
				case 8:
					imgbuf[idxnow++] = buffer[x + offset];
					break;
				}
			}

			if(m_bitsperpixel != 8)
			{
				tabidx *= 3;
				imgbuf[idxnow++] = m_table[tabidx + 2];
				imgbuf[idxnow++] = m_table[tabidx + 1];
				imgbuf[idxnow++] = m_table[tabidx];
			}
		}
	}
	delete buffer;
}

/////////////////////////////////////////////////////////////////////////////
// Decompress8()
//
// Description: Decompress 8bit RLE image.
//
// TSTan   20 Oct 2002
/////////////////////////////////////////////////////////////////////////////
void CPCXEngine::Decompress8()
{
	if(m_nplanes == 1)	// 8-bit
	{
		// Read img to the end of buffer
		int idx;
		idx = 2 * GetX() * GetY();

		char *srcimg = &m_image[idx];
		Decompress(srcimg);
		// Read palette
		if(m_table)
			delete m_table;
		int colors = 1 << m_bitsperpixel;
		m_table = new char[3 * colors];

		if(m_iobits->GetBits(8) != 12)
			return;

		int clridx;
		for(clridx = 0; clridx < colors; clridx++)
		{
			m_table[3 * clridx] = m_iobits->GetBits(8);
			m_table[3 * clridx + 1] = m_iobits->GetBits(8);
			m_table[3 * clridx + 2] = m_iobits->GetBits(8);
		}
		// Reformat buffer
		idx >>= 1;
		char *ptrimg = m_image;
		int tabidx;
		for(clridx = 0; clridx < idx; clridx++)
		{
			tabidx = 3 * (*srcimg & 0x00ff);
			*ptrimg = m_table[tabidx + 0];
			ptrimg++;
			*ptrimg = m_table[tabidx + 1];
			ptrimg++;
			*ptrimg = m_table[tabidx + 2];
			ptrimg++;
			srcimg++;
		}
	}
	else if(m_nplanes == 3)	// 24-bit
		Decompress(m_image);
}

/////////////////////////////////////////////////////////////////////////////
// GetImage()
//
// Description: Return pointer to image RGB array.
//
// TSTan   20 Oct 2002
/////////////////////////////////////////////////////////////////////////////
char *CPCXEngine::GetImage()
{
	return m_image;
}

/////////////////////////////////////////////////////////////////////////////
// GetX()
//
// Description: Return width of image.
//
// TSTan   20 Oct 2002
/////////////////////////////////////////////////////////////////////////////
int CPCXEngine::GetX()
{
	return m_xmax - m_xmin + 1;
}

/////////////////////////////////////////////////////////////////////////////
// GetY()
//
// Description: Return height of image.
//
// TSTan   20 Oct 2002
/////////////////////////////////////////////////////////////////////////////
int CPCXEngine::GetY()
{
	return m_ymax - m_ymin + 1;
}

/////////////////////////////////////////////////////////////////////////////
// LittleEnd(int data, int bytes)
//
// Description: Reverse data bytes to form Little Endian word.
//
// data - data in Big Endian.
// bytes - length of data in bytes.
//
// TSTan   20 Oct 2002
/////////////////////////////////////////////////////////////////////////////
int CPCXEngine::LittleEnd(int data, int bytes)
{
	int retdata = 0;
	while(bytes--)
	{
		retdata <<= 8;
		retdata |= data & 0x00ff;
		data >>= 8;
	}

	return retdata;
}

/////////////////////////////////////////////////////////////////////////////
// ProcessInFile(char *filename)
//
// Description: Decode the PCX file.
//
// filename - PCX file name.
//
// TSTan   20 Oct 2002
/////////////////////////////////////////////////////////////////////////////
void CPCXEngine::ProcessInFile(char *filename)
{
	Reset();

	CFileBitStream iobitstream(filename);
	m_iobits = &iobitstream;

	if(iobitstream.RdBytesAvailable())
	{
/*
ZSoft .PCX FILE HEADER FORMAT

Byte      Item          Size   Description/Comments 
 0         Manufacturer 1      Constant Flag, 10 = ZSoft .pcx 
 1         Version      1      Version information 
                               0 = Version 2.5 of PC Paintbrush 
                               2 = Version 2.8 w/palette information 
                               3 = Version 2.8 w/o palette information 
                               4 = PC Paintbrush for Windows(Plus for
                                  Windows uses Ver 5) 
                               5 = Version 3.0 and > of PC Paintbrush
                                  and PC Paintbrush +, includes
                                  Publisher's Paintbrush . Includes
                                  24-bit .PCX files 
 2         Encoding      1     1 = .PCX run length encoding 
 3         BitsPerPixel  1     Number of bits to represent a pixel
                                  (per Plane) - 1, 2, 4, or 8 
 4         Window        8     Image Dimensions: Xmin,Ymin,Xmax,Ymax 
12         HDpi          2     Horizontal Resolution of image in DPI* 
14         VDpi          2     Vertical Resolution of image in DPI* 
16         Colormap     48     Color palette setting, see text 
64         Reserved      1     Should be set to 0. 
65         NPlanes       1     Number of color planes 
66         BytesPerLine  2     Number of bytes to allocate for a scanline
                                  plane.  MUST be an EVEN number.  Do NOT
                                  calculate from Xmax-Xmin. 
68         PaletteInfo   2     How to interpret palette- 1 = Color/BW,
                                  2 = Grayscale (ignored in PB IV/ IV +) 
70         HscreenSize   2     Horizontal screen size in pixels. New field
                                  found only in PB IV/IV Plus 
72         VscreenSize   2     Vertical screen size in pixels. New field
                                  found only in PB IV/IV Plus 
74         Filler       54     Blank to fill out 128 byte header.  Set all
                                  bytes to 0 
*/
		// Get header
		m_manufacturer = m_iobits->GetBits(8);
		m_version = m_iobits->GetBits(8);
		m_encoding = m_iobits->GetBits(8);
		m_bitsperpixel = m_iobits->GetBits(8);
		m_xmin = LittleEnd(m_iobits->GetBits(16), 2);
		m_ymin = LittleEnd(m_iobits->GetBits(16), 2);
		m_xmax = LittleEnd(m_iobits->GetBits(16), 2);
		m_ymax = LittleEnd(m_iobits->GetBits(16), 2);
		m_hdpi = LittleEnd(m_iobits->GetBits(16), 2);
		m_vdpi = LittleEnd(m_iobits->GetBits(16), 2);

		if(m_table)
			delete m_table;
		m_table = new char[48];
		m_iobits->FillArray(m_table, 48);
		if(m_iobits->GetBits(8) != 0)
			return;

		m_nplanes = m_iobits->GetBits(8);
		m_bytesperline = LittleEnd(m_iobits->GetBits(16), 2);
		m_paletteinfo = LittleEnd(m_iobits->GetBits(16), 2);
		m_hscreensize = LittleEnd(m_iobits->GetBits(16), 2);
		m_vscreensize = LittleEnd(m_iobits->GetBits(16), 2);

		int idx;
		for(idx = 0; idx < 54; idx++)
		{
			if(m_iobits->GetBits(8))
				return;
		}

		int width = m_xmax - m_xmin + 1;
		int height = m_ymax - m_ymin + 1;
		int imgW = 3 * width;
		if(imgW % 4)
			imgW += 4 - (imgW % 4);
		m_image = new char[imgW * height];
		
		if(m_bitsperpixel == 8)
			Decompress8();
		else
			Decompress(m_image);
	}
}

/////////////////////////////////////////////////////////////////////////////
// Reset()
//
// Description: Reset to prepare for new image.
//
// TSTan   20 Oct 2002
/////////////////////////////////////////////////////////////////////////////
void CPCXEngine::Reset()
{
	if(m_image)
		delete []m_image;
	m_image = NULL;

	if(m_table)
		delete []m_table;
	m_table = NULL;
}

/////////////////////////////////////////////////////////////////////////////
// SetNewImage(char *img, int x, int y)
//
// Description: Set buffer to new RGB buffer.
//
// img - char array holding new image's RGB.
// x - X dimension.
// y - Y dimension.
//
// TSTan   10 Nov 2002
/////////////////////////////////////////////////////////////////////////////
void CPCXEngine::SetNewImage(char *img, int x, int y)
{
	if(m_image)
		delete []m_image;
	m_image = img;
	m_xmin = m_ymin = 0;
	m_xmax = x - 1;
	m_ymax = y - 1;
}
