///////////////////Software Control Header///////////////////////////////////
// TIFFLib.cpp - Library to support TIFF encoding/decoding.
//
// $LogThisFile$ \Projects\UtilCode\Lib\TIFFLib.cpp
// 
// V01.01 30/11/2002 0242
//  Format image buffer to Windows DIB format.
// 
// V01.00 03/11/2002 0018
//  Initial revision.
// 
// This code is freely distributed to help the development of complex
// software in any field using C++. The author is not held responsible
// for any damage caused by the usage of the code.
// The author should be acknowledged for the use of this piece of code in
// commercial software.
//
// Bug reports and suggestions are welcomed to report to
//
//          tzzstan@hotmail.com
/////////////////////////////////////////////////////////////////////////////
#include "TIFFLib.h"

/////////////////////////////////////////////////////////////////////////////
// CTIFF_LZWCode()
//
// Description: LZW code word object constructor.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
CTIFF_LZWCode::CTIFF_LZWCode()
{
	m_codes = NULL;
}

/////////////////////////////////////////////////////////////////////////////
// ~CTIFF_LZWCode()
//
// Description: Cleanup.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
CTIFF_LZWCode::~CTIFF_LZWCode()
{
	Reset();
}

/////////////////////////////////////////////////////////////////////////////
// AddCode(int code)
//
// Description: Add a code into code word.
//
// code - code to add.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
void CTIFF_LZWCode::AddCode(int code)
{
	if(m_codes == NULL)
	{
		m_codes = new char;
		m_codelength = 1;
		m_codes[0] = code;
	}
	else
	{
		char *tmpcodes = new char[m_codelength + 1];
		int idx;
		for(idx = 0; idx < m_codelength; idx++)
			tmpcodes[idx] = m_codes[idx];
		tmpcodes[m_codelength] = code;
		m_codelength++;

		delete []m_codes;
		m_codes = tmpcodes;
	}
}

/////////////////////////////////////////////////////////////////////////////
// AddCodes(CTIFF_LZWCode &code)
//
// Description: Append another object's code word.
//
// code - the source object.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
void CTIFF_LZWCode::AddCodes(CTIFF_LZWCode &code)
{
	int idx;
	char *codeptr;

	if(m_codes == NULL)
	{
		m_codes = new char[code.m_codelength];
		m_codelength = code.m_codelength;
		codeptr = m_codes;
	}
	else
	{
		char *tmpcodes = new char[m_codelength + code.m_codelength];
		for(idx = 0; idx < m_codelength; idx++)
			tmpcodes[idx] = m_codes[idx];

		delete m_codes;
		m_codes = tmpcodes;
		codeptr = &m_codes[m_codelength];
		m_codelength += code.m_codelength;
	}

	for(idx = 0; idx < code.m_codelength; idx++)
	{
		*codeptr = code.m_codes[idx];
		codeptr++;
	}
}

/////////////////////////////////////////////////////////////////////////////
// Reset()
//
// Description: Reset to prepare for new code word.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
void CTIFF_LZWCode::Reset()
{
	if(m_codes)
		delete []m_codes;
	m_codes = NULL;
}

/////////////////////////////////////////////////////////////////////////////
// CTIFF_IFDEntry()
//
// Description: Constructor for an IFD object.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
CTIFF_IFDEntry::CTIFF_IFDEntry()
{
	m_data = NULL;
}

/////////////////////////////////////////////////////////////////////////////
// ~CTIFF_IFDEntry()
//
// Description: Cleanup.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
CTIFF_IFDEntry::~CTIFF_IFDEntry()
{
	if(m_data)
		delete m_data;
}

/////////////////////////////////////////////////////////////////////////////
// CTIFFEngine()
//
// Description: Contructor to initialize pointers.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
CTIFFEngine::CTIFFEngine()
{
	m_table = NULL;
	m_image = NULL;
	m_LZWCode = NULL;
}

/////////////////////////////////////////////////////////////////////////////
// ~CTIFFEngine()
//
// Description: Cleanup.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
CTIFFEngine::~CTIFFEngine()
{
	Reset();
}

/////////////////////////////////////////////////////////////////////////////
// ByteInGet(int arg)
//
// Description: Return a byte from image with approriate decompression.
//
// arg - current x coordinate.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
char CTIFFEngine::ByteInGet(int arg)
{
	char retvalue;

	if(m_compression == 32773)
	{
		if(!m_control)
		{
			m_control = m_iobits->GetBits(8);
			if(m_control >= 0 && m_control <= 127)
			{
				m_count = m_control + 1;
				m_control = 1;
			}
			else if(m_control != -128)
			{
				m_count = -m_control + 1;
				m_data = m_iobits->GetBits(8);
				m_control = 2;
			}
			else
			{
				m_control = 0;
				return ByteInGet();
			}
		}

		if(m_control == 1)
			retvalue = m_iobits->GetBits(8);
		else
			retvalue = m_data;

		m_count--;
		if(!m_count)
			m_control = 0;
	}
	else if(m_compression == 2)	// CCITT - Huffman
	{
		// The following 2 lines are the only differences compared
		// to compression - 3
		if(!arg)
			m_iobits->SkipBits(m_iobits->RemainingBits() % 8);
		//////////////////////////////////////////////////////////

		int bits = m_imagewidth - arg;
		if(bits > 8)
			bits = 8;

		retvalue = 0;
		int idx;
		for(idx = 0; idx < bits; idx++)
		{
			if(!m_count)	// Decompress
			{
				m_data = 1 - m_data;	// Alternating black and white

				// Get count using Make-up and Terminating tables
				m_count = GetLength();
			}

			if(!m_count)	// Possible of 0 count
			{
				idx--;
				continue;
			}
			else
			{
				retvalue <<= 1;
				retvalue |= m_data;
				m_count--;
			}
		}

		if(m_photometricinterpretation)
			retvalue = ~retvalue;

		retvalue <<= 8 - bits;
	}
	else if(m_compression == 3)	// CCITT
	{
		int bits = m_imagewidth - arg;
		if(bits > 8)
			bits = 8;

		retvalue = 0;
		int idx;
		for(idx = 0; idx < bits; idx++)
		{
			if(!m_count)	// Decompress
			{
				m_data = 1 - m_data;	// Alternating black and white

				// Get count using Make-up and Terminating tables
				m_count = GetLength();
			}

			if(!m_count)	// Possible of 0 count
			{
				idx--;
				continue;
			}
			else
			{
				retvalue <<= 1;
				retvalue |= m_data;
				m_count--;
			}
		}

		if(m_photometricinterpretation)
			retvalue = ~retvalue;

		retvalue <<= 8 - bits;
	}
	else if(m_compression == 5)	// LZW
	{
		int code, idx;

		if(!m_lzw_count)	// Get new codes
		{
			code = m_iobits->GetBits(m_lzw_codelength);
			if(code == 256)	// ClearCode
			{
				// Initialize
				for(idx = 0; idx < 4096; idx++)
				{
					m_LZWCode[idx].Reset();
					if(idx < 256)
						m_LZWCode[idx].AddCode(idx);
					m_lzw_codelength = 9;
				}
				code = m_iobits->GetBits(m_lzw_codelength);
				if(code == 257)	// EOI
					return 0;
				m_lzw_ptr = &m_LZWCode[code];
				m_lzw_count = m_lzw_ptr->m_codelength;
				m_lzw_oldcode = code;
				m_lzw_filling = 258;
			}
			else
			{
				if(m_LZWCode[code].m_codes)	// In table
				{
					m_lzw_ptr = &m_LZWCode[code];
					m_LZWCode[m_lzw_filling].AddCodes(m_LZWCode[m_lzw_oldcode]);
					m_LZWCode[m_lzw_filling].AddCode(m_lzw_ptr->m_codes[0]);
				}
				else
				{
					// Add string to table
					m_lzw_ptr = &m_LZWCode[m_lzw_filling];
					m_lzw_ptr->AddCodes(m_LZWCode[m_lzw_oldcode]);
					m_lzw_ptr->AddCode(m_LZWCode[m_lzw_oldcode].m_codes[0]);
				}
				m_lzw_count = m_lzw_ptr->m_codelength;
				m_lzw_oldcode = code;

				switch(m_lzw_filling)
				{
				case 510:
				case 1022:
				case 2046:
					m_lzw_codelength++;
				}
				m_lzw_filling++;
			}
		}

		retvalue = m_lzw_ptr->m_codes[m_lzw_ptr->m_codelength - m_lzw_count];
		m_lzw_count--;
	}
	else
		retvalue = m_iobits->GetBits(8);;

	return retvalue;
}

/////////////////////////////////////////////////////////////////////////////
// ByteInReset()
//
// Description: Prepare for new scan line.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
void CTIFFEngine::ByteInReset()
{
	m_control = 0;
	m_count = 0;
	m_data = 1;
}

/////////////////////////////////////////////////////////////////////////////
// DecodeImage()
//
// Description: Decode the image into RGB array.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
void CTIFFEngine::DecodeImage()
{
	int x, y, data, red, green, blue, coloridx, oldred, oldgreen, oldblue;
	char *stripoffset = m_stripoffsets;
	char *ptrimg;

	if(m_compression == 5)
		m_LZWCode = new CTIFF_LZWCode[4096];

	int imgW = 3 * m_imagewidth;
	if(imgW % 4)
		imgW += 4 - (imgW % 4);

	for(y = 0; y < m_imagelength; y++)
	{
		if(!(y % m_rowsperstrip))
		{
			m_iobits->Seek(GetEndianWord(stripoffset, 4));
			stripoffset++;
			stripoffset++;
			stripoffset++;
			stripoffset++;
			m_lzw_codelength = 9;
			m_lzw_count = 0;
		}

		ptrimg = &m_image[y * imgW];
		ByteInReset();
		oldred = oldgreen = oldblue = 0;
		for(x = 0; x < m_imagewidth; x++)
		{
			// Bilevel
			if(m_bitspersample == 1)
			{
				if(!(x % 8))
					data = ByteInGet(x);

				if(data & (0x0080 >> (x % 8)))
				{
					red = m_photometricinterpretation? 0x00ff: 0;
					blue = green = red;
				}
				else
				{
					red = m_photometricinterpretation? 0: 0x00ff;
					blue = green = red;
				}
			}
			// 8/4bit image
			else if(m_bitspersample <= 8)
			{
				if(!(x % (8 / m_bitspersample)))
				{
					data = ByteInGet(x);
					coloridx = data & 0xff;
					if(m_bitspersample == 4)
						coloridx = (data >> 4) & 0xf;
				}
				else
					coloridx = data & 0xf;

				coloridx *= 3;
				red = m_table[coloridx++];
				green = m_table[coloridx++];
				blue = m_table[coloridx++];
			}
			// 24bit image
			else if(m_bitspersample == 24)
			{
				red = ByteInGet(x);
				green = ByteInGet(x);
				blue = ByteInGet(x);
			}

			if(m_predictor == 2)	// Differences
			{
				red += oldred;
				green += oldgreen;
				blue += oldblue;

				oldred = red;
				oldgreen = green;
				oldblue = blue;
			}

			*ptrimg = blue;
			ptrimg++;
			*ptrimg = green;
			ptrimg++;
			*ptrimg = red;
			ptrimg++;
		}
	}

	if(m_LZWCode)
	{
		delete []m_LZWCode;
		m_LZWCode = NULL;
	}
}

/////////////////////////////////////////////////////////////////////////////
// Endian(int data, int bytes)
//
// Description: Reverse data bytes to form Little Endian word based on
//              m_order IFD of TIFF.
//
// data - data in Big/Little Endian.
// bytes - length of data in bytes.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
int CTIFFEngine::Endian(int data, int bytes)
{
	if(m_order == 0x4d4d)
		return data;

	int retdata = 0;
	while(bytes--)
	{
		retdata <<= 8;
		retdata |= data & 0x00ff;
		data >>= 8;
	}

	return retdata;
}

/////////////////////////////////////////////////////////////////////////////
// GetEndianWord(char *ptr, int bytes)
//
// Description: Prepare a word from array (could be Big/Little Endian).
//
// ptr - pointer to the array.
// bytes - length of word in bytes.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
int CTIFFEngine::GetEndianWord(char *ptr, int bytes)
{
	int value = 0;
	int oldbytes = bytes;

	while(bytes--)
	{
		value <<= 8;
		value |= (*ptr & 0xff);
		ptr++;
	}

	return Endian(value, oldbytes);
}

/////////////////////////////////////////////////////////////////////////////
// GetImage()
//
// Description: Return pointer to image RGB array.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
char *CTIFFEngine::GetImage()
{
	return m_image;
}

/////////////////////////////////////////////////////////////////////////////
// GetLength()
//
// Description: Return length of next data using Modified Huffman
//              decompression.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
int CTIFFEngine::GetLength()
{
	int length = 0;
	int codelength = 0;
	int code = 0;

	const int whitecode[] =
		{0x07, 0x08, 0x0b, 0x0c, 0x0e, 0x0f, 0x13, 0x14, 0x07, 0x08,
		 0x1b, 0x12, 0x07, 0x08, 0x03, 0x34, 0x35, 0x2a, 0x2b, 0x17,
		 0x18, 0x27, 0x0c, 0x08, 0x17, 0x03, 0x04, 0x28, 0x2b, 0x13,
		 0x24, 0x18, 0x37, 0x35, 0x02, 0x03, 0x1a, 0x1b, 0x12, 0x13,
		 0x14, 0x15, 0x16, 0x17, 0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d,
		 0x04, 0x05, 0x0a, 0x0b, 0x52, 0x53, 0x54, 0x55, 0x24, 0x25,
		 0x58, 0x59, 0x5a, 0x5b, 0x4a, 0x4b, 0x32, 0x33, 0x34, 0x36,
		 0x37, 0x64, 0x65, 0x68, 0x67, 0xcc, 0xcd, 0xd2, 0xd3, 0xd4,
		 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xdb, 0x98, 0x99, 0x9a,
		 0x9b, 0x08, 0x0c, 0x0d, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
		 0x1c, 0x1d, 0x1e, 0x1f, 0x01};
	const int blackcode[] =
		{0x03, 0x02, 0x02, 0x03, 0x03, 0x02, 0x03, 0x05, 0x04, 0x04,
		 0x05, 0x07, 0x04, 0x07, 0x18, 0x37, 0x17, 0x18, 0x08, 0x0f,
		 0x67, 0x68, 0x6c, 0x37, 0x28, 0x17, 0x18, 0x08, 0x0c, 0x0d, 0x00,
		 0xca, 0xcb, 0xcc, 0xcd, 0x68, 0x69, 0x6a, 0x6b, 0xd2, 0xd3,
		 0xd4, 0xd5, 0xd6, 0xd7, 0x6c, 0x6d, 0xda, 0xdb, 0x54, 0x55,
		 0x56, 0x57, 0x64, 0x65, 0x52, 0x53, 0x24, 0x37, 0x38, 0x27,
		 0x28, 0x58, 0x59, 0x2b, 0x2c, 0x5a, 0x66, 0x67, 0xc8, 0xc9,
		 0x5b, 0x33, 0x34, 0x35, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
		 0x1c, 0x1d, 0x1e, 0x1f, 0x6c, 0x6d, 0x4a, 0x4b, 0x4c, 0x4d,
		 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x52, 0x53, 0x54, 0x55,
		 0x5a, 0x5b, 0x64, 0x65};
	const int whitelength[] =
		{   2,    3,    4,    5,    6,    7,    8,    9,   10,   11,
		   64,  128,    1,   12,   13,   14,   15,   16,   17,  192,
		 1664,   18,   19,   20,   21,   22,   23,   24,   25,   26,
		   27,   28,  256,    0,   29,   30,   31,   32,   33,   34,
		   35,   36,   37,   38,   39,   40,   41,   42,   43,   44,
		   45,   46,   47,   48,   49,   50,   51,   52,   53,   54,
		   55,   56,   57,   58,   59,   60,   61,   62,   63,  320,
		  384,  448,  512,  576,  640,  704,  768,  832,  896,  960,
		 1024, 1088, 1152, 1216, 1280, 1344, 1408, 1472, 1536, 1600,
		 1728, 1792, 1856, 1920, 1984, 2048, 2112, 2176, 2240, 2304,
		 2368, 2432, 2496, 2560, 3000};
	const int blacklength[] = 
		{   2,    3,    1,    4,    5,    6,    7,    8,    9,   10,
		   11,   12,   13,   14,   15,    0,   16,   17,   18,   64,
		   19,   20,   21,   22,   23,   24,   25, 1792, 1856, 1920, 3000,
		   26,   27,   28,   29,   30,   31,   32,   33,   34,   35,
		   36,   37,   38,   39,   40,   41,   42,   43,   44,   45,
		   46,   47,   48,   49,   50,   51,   52,   53,   54,   55,
		   56,   57,   58,   59,   60,   61,   62,   63,  128,  192,
		  256,  320,  384,  448, 1984, 2048, 2112, 2176, 2240, 2304,
		 2368, 2432, 2496, 2560,  512,  576,  640,  704,  768,  832,
		  896,  960, 1024, 1088, 1152, 1216, 1280, 1344, 1408, 1472,
		 1536, 1600, 1664, 1728};
	const int whitestart[] = {0, 0, 0, 0, 0, 6, 12, 21, 33, 75, 75,
		91, 104, 105, 105};
	const int blackstart[] = {0, 0, 0, 2, 4, 6,  7,  9, 12, 14, 15,
		20, 31, 86, 105};

	const int *ptrcode, *ptrlength, *ptrstart;
	if(m_data)	// black
	{
		ptrcode = blackcode;
		ptrlength = blacklength;
		ptrstart = blackstart;
	}
	else
	{
		ptrcode = whitecode;
		ptrlength = whitelength;
		ptrstart = whitestart;
	}

	int idx;
	while(1)
	{
		code <<= 1;
		code |= m_iobits->GetBits(1);
		codelength++;

		if(codelength > 13)	// Error
			return 0;

		for(idx = ptrstart[codelength]; idx < ptrstart[codelength + 1]; idx++)
		{
			if(ptrcode[idx] == code)	// Found Huffman code
			{
				if(ptrlength[idx] < 3000)
					length += ptrlength[idx];
				else
				{
					codelength = 0;
					code = 0;
					break;
				}
				if(ptrlength[idx] < 64)
					return length;
				else
				{
					codelength = 0;
					code = 0;
					break;
				}
			}
		}
	}

	return length;
}

/////////////////////////////////////////////////////////////////////////////
// GetX()
//
// Description: Return width of image.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
int CTIFFEngine::GetX()
{
	return m_imagewidth;
}

/////////////////////////////////////////////////////////////////////////////
// GetY()
//
// Description: Return height of image.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
int CTIFFEngine::GetY()
{
	return m_imagelength;
}

/////////////////////////////////////////////////////////////////////////////
// ProcessIFDs()
//
// Description: Decipher the IFD messages.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
void CTIFFEngine::ProcessIFDs()
{
	int totalcolor, idx, triplet, colorcomp;
	char *src, *ptr;

	for(m_ifdentries.ResetPtr(); m_ifdentries.IsValid(); m_ifdentries.Next())
	{
		CTIFF_IFDEntry *ifd = m_ifdentries.Current();
		switch(ifd->m_tag)
		{
		// Compression
		case 0x0103:
			m_compression = ifd->m_offset;
			break;
		// ImageWidth
		case 0x100:
			m_imagewidth = ifd->m_offset;
			break;
		// ImageLength
		case 0x101:
			m_imagelength = ifd->m_offset;
			break;
		// RowsPerStrip
		case 0x116:
			m_rowsperstrip = ifd->m_offset;
			break;
		// StripOffsets
		case 0x111:
			if(ifd->m_data == NULL)
			{
				ifd->m_data = new char[4];
				char *ptr = (char *)&ifd->m_offset;
				(ifd->m_data)[0] = *ptr;
				*ptr++;
				(ifd->m_data)[1] = *ptr;
				*ptr++;
				(ifd->m_data)[2] = *ptr;
				*ptr++;
				(ifd->m_data)[3] = *ptr;
			}
			m_stripoffsets = ifd->m_data;
			break;
		// BitsPerSample
		case 0x102:
			m_bitspersample = ifd->m_offset;
			break;
		// PhotometricInterpretation
		// 0 = WhiteIsZero.
		// 1 = BlackIsZero.
		case 0x106:
			m_photometricinterpretation = ifd->m_offset;
			break;
		// ColorMap
		case 0x140:
			if(m_photometricinterpretation != 3)
				break;
			totalcolor = 1 << m_bitspersample;
			m_table = new char[3 * totalcolor];
			ptr = m_table;
			src = ifd->m_data;
			for(triplet = 0; triplet < 3; triplet++)
			{
				for(idx = 0; idx < totalcolor; idx++)
				{
					colorcomp = *((int*)src);
					m_table[3 * idx + triplet] =
						(Endian(colorcomp, 2) >> 8) & 0xff;
					src++;
					src++;
				}
			}
			break;
		// SamplesPerPixel
		case 0x115:
			m_samplesperpixel = ifd->m_offset;
			break;
		// Predictor
		case 0x13d:
			m_predictor = ifd->m_offset;
			break;
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
// ProcessInFile(char *filename)
//
// Description: Decode the TIFF file.
//
// filename - TIFF file name.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
void CTIFFEngine::ProcessInFile(char *filename)
{
	Reset();

	CFileBitStream iobitstream(filename);
	m_iobits = &iobitstream;

	if(iobitstream.RdBytesAvailable())
	{
		// Get header
		m_order = m_iobits->GetBits(16);
		if(Endian(m_iobits->GetBits(16), 2) != 0x2a)
			return;
		m_offset = Endian(m_iobits->GetBits(32), 4);
		m_iobits->Seek(m_offset);

		int entries = Endian(m_iobits->GetBits(16), 2);

		// Support only 1 IFD
		int idx;
		for(idx = 0; idx < entries; idx++)
		{
			CTIFF_IFDEntry *ifd = new CTIFF_IFDEntry;
			ifd->m_tag = Endian(m_iobits->GetBits(16), 2);
			ifd->m_type = Endian(m_iobits->GetBits(16), 2);
			ifd->m_count = Endian(m_iobits->GetBits(32), 4);
			ifd->m_offset = Endian(m_iobits->GetBits(32), 4);

			// Sequence is important, insert at tail.
			m_ifdentries.Insert(ifd, CSinglyList<CTIFF_IFDEntry>::TAIL);

// 1 = BYTE 8-bit unsigned integer.
// 2 = ASCII 8-bit byte that contains a 7-bit ASCII code; the last byte
//     must be NUL (binary zero).
// 3 = SHORT 16-bit (2-byte) unsigned integer.
// 4 = LONG 32-bit (4-byte) unsigned integer.
// 5 = RATIONAL Two LONGs: the first represents the numerator of a
//     fraction; the second, the denominator.
			int singleentry, totalbytes;
			switch(ifd->m_type)
			{
			case 1:
			case 2:
				singleentry = 1;
				break;
			case 3:
				singleentry = 2;
				break;
			case 4:
				singleentry = 4;
				break;
			case 5:
				singleentry = 8;
				break;
			}

			totalbytes = singleentry * ifd->m_count;
			if(totalbytes > 4)
			{
				int offset = m_iobits->GetOffset();
				m_iobits->Seek(ifd->m_offset);
				ifd->m_data = new char[totalbytes];
				m_iobits->FillArray(ifd->m_data, totalbytes);
				m_iobits->Seek(offset);
			}
		}
		ProcessIFDs();

		int imgW = 3 * m_imagewidth;
		if(imgW % 4)
			imgW += 4 - (imgW % 4);

		m_image = new char[imgW * m_imagelength];
		DecodeImage();
	}
}

/////////////////////////////////////////////////////////////////////////////
// Reset()
//
// Description: Reset to prepare for new image.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
void CTIFFEngine::Reset()
{
	if(m_image)
		delete m_image;
	m_image = NULL;

	if(m_table)
		delete m_table;
	m_table = NULL;

	m_ifdentries.RemoveAll();
	m_photometricinterpretation = -1;
	m_predictor = 1;
}
