///////////////////Software Control Header///////////////////////////////////
// TString.cpp - Library to support string manipulation
//
// $LogThisFile$ \Projects\UtilCode\Lib\TString.cpp
// 
// V01.05 29/06/2002 1442
//  Scientific notation display.
//  Functions supporting Replace() added.
// 
// V01.04 17/06/2001 1959
//  TrimRight and TrimLeft added.
// 
// V01.03 14/03/2001 2135
//  ToInteger(int Location) added.
// 
// V01.02 29/10/2000 2315
//  RemoveLeft() and operator+=(float) added.
// 
// V01.01 26/10/2000 2155
//  Various changes to deallocate memory only if exceeding m_Limit.
//  Default to 1024 bytes.
// 
// V01.00 11/10/2000 2141
//  Initial revision.
// 
// This code is freely distributed to help the development of complex
// software in any field using C++. The author is not held responsible
// for any damage caused by the usage of the code.
// The author should be acknowledged for the use of this piece of code in
// commercial software.
//
// Bug reports and suggestions are welcomed to report to
//
//          tzzstan@hotmail.com
/////////////////////////////////////////////////////////////////////////////
#include <string.h>

#include "TString.h"

// Instantiate static members
CTString CTString::m_TempStorage;

/////////////////////////////////////////////////////////////////////////////
// CTString(char *string)
//
// Description: Construct from char*.
//
// string - Pointer to string kept in char array.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
CTString::CTString(char *string, int LimitingSize)
{
	m_flag = str_scientific;
	m_precision = 2;
	m_width = 0;
   m_Limit = LimitingSize;
   m_Current = strlen(string) + 1;
   m_data = new char[m_Current];
   strcpy(m_data, string);
}

/////////////////////////////////////////////////////////////////////////////
// CTString(const CTString &string)
//
// Description: Copy construct from string.
//
// string - Reference to another CTString object.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
CTString::CTString(const CTString &string)
{
	m_flag = string.m_flag;
	m_precision = string.m_precision;
	m_width = string.m_width;
   m_Limit = string.m_Limit;
   m_Current = string.m_Current;
   m_data = new char[m_Current];
   strcpy(m_data, string.m_data);
}

/////////////////////////////////////////////////////////////////////////////
// ~CTString()
//
// Description: Cleanup.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
CTString::~CTString()
{
   delete []m_data;
}

/////////////////////////////////////////////////////////////////////////////
// DeleteLine(int linenumber)
//
// Description: Delete a line from m_data based on '\n' or '\r'.
//
// linenumber - line number to delete.
//
// TSTan   29 Jun 2002
/////////////////////////////////////////////////////////////////////////////
void CTString::DeleteLine(int linenumber)
{
	int idx = 0;
	int linecount = -1;
	int startidx = 0;
	int oldline = -1;

	while(m_data[idx])
	{
		if(m_data[idx] == '\n')
		{
			if(m_data[idx + 1] == '\r')
				idx++;
			linecount++;
		}
		else if(m_data[idx] == '\r')
		{
			if(m_data[idx + 1] == '\n')
				idx++;
			linecount++;
		}

		if(linecount == linenumber)
			break;

		if(oldline != linecount)
			startidx = idx + 1;

		oldline = linecount;
		idx++;
	}

	if(linecount < 0)
		linecount++;

	if(linecount == linenumber)
		Removes(startidx, idx - startidx + 1);
}

/////////////////////////////////////////////////////////////////////////////
// Find(char *string, int index)
//
// Description: Find substring within string starting from index.
//              Return index to the first character. -1 if not found.
//
// string - substring to search for.
// index - index within m_data to start searching.
//
// TSTan   29 Jun 2002
/////////////////////////////////////////////////////////////////////////////
int CTString::Find(char *string, unsigned int index)
{
	if(index >= strlen(m_data))
		return -1;

	int charcount;
	bool found;

	while(m_data[index])
	{
		charcount = 0;
		found = true;
		while(string[charcount])
		{
			if(m_data[index + charcount] != string[charcount])
			{
				found = false;
				break;
			}
			charcount++;
		}
		if(found)
			return index;

		index++;
	}
	return -1;
}

/////////////////////////////////////////////////////////////////////////////
// GetChar(char Character, int FromLeftRight)
//
// Description: Find Character within string. Return its offset.
//
// Character - Character to search for.
// FromLeftRight - 0: Search from left.
//                 1: Search from right.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
int CTString::GetChar(char Character, int FromLeftRight)
{
   int Length = Size();
   int Position;

   if(FromLeftRight == 0)  // Search from left
   {
      for(Position = 0; Position < Length; Position++)
      {
         if(m_data[Position] == Character)
            return Position;
      }
   }
   else  // Search from right
   {
      for(Position = Length - 1; Position >= 0; Position--)
      {
         if(m_data[Position] == Character)
            return Position;
      }
   }
   return -1;  // Cannot find
}

/////////////////////////////////////////////////////////////////////////////
// InsertAt(int Location, char Character)
//
// Description: Insert Character at offset Location.
//
// Location - Offset to be inserted Character.
// Character - Character to insert.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
void CTString::InsertAt(int Location, char Character)
{
   int Length = strlen(m_data);
   char *t_ptr;

   if(Location > Length)   // Out of range index
      return;

   if(m_Current < Length + 2)
   {
      m_Current = Length + 2;
      t_ptr = new char[m_Current];
      strcpy(t_ptr, m_data);
      delete []m_data;
      m_data = t_ptr;
   }

   int CurIdx = Location;
   char tempChar = m_data[CurIdx];
   char tempChar2;
   m_data[CurIdx] = Character;
   while(tempChar)
   {
      tempChar2 = m_data[++CurIdx];
      m_data[CurIdx] = tempChar;
      tempChar = tempChar2;
   }
   m_data[++CurIdx] = tempChar;
}

/////////////////////////////////////////////////////////////////////////////
// InsertAt(int Location, CTString &string)
//
// Description: Insert string at offset Location.
//
// Location - Offset to insert string.
// string - string to be inserted.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
void CTString::InsertAt(int Location, CTString &string)
{
	InsertAt(Location, string.m_data);
}

/////////////////////////////////////////////////////////////////////////////
// InsertAt(int Location, char *string)
//
// Description: Insert string at offset Location.
//
// Location - Offset to insert string.
// string - string to be inserted.
//
// TSTan   29 Jun 2000
/////////////////////////////////////////////////////////////////////////////
void CTString::InsertAt(int Location, char *string)
{
   int Length = strlen(string);
   int Position;

   for(Position = Length - 1; Position >= 0; Position--)
      InsertAt(Location, string[Position]);
}

/////////////////////////////////////////////////////////////////////////////
// Lines()
//
// Description: Line count based on '\n'
//
// TSTan   29 Jun 2002
/////////////////////////////////////////////////////////////////////////////
int CTString::Lines()
{
	int linecount = 0;
	int idx = 0;

	while(m_data[idx])
	{
		if(m_data[idx++] == '\n')
			linecount++;
	}
	
	return linecount;
}

/////////////////////////////////////////////////////////////////////////////
// RemoveAt(int Location)
//
// Description: Remove one character from string.
//
// Location - Offset to the character to be removed.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
int CTString::RemoveAt(int Location)
{
   int Length = strlen(m_data);
   int Removed = 0;
   char *t_ptr = m_data;
   
   if(Location < Length)
   {
      int Position = 0;
      char *L_ptr = t_ptr;
      while(m_data[Position])
      {
         *L_ptr = m_data[Position];
         if(Position != Location)
            L_ptr++;
         else
            Removed = *L_ptr;
         Position++;
      }
      *L_ptr = '\0';
   }
   else
      return 0;

   if(m_Current > m_Limit)
   {
      m_Current = Length;
      t_ptr = new char[m_Current];
      strcpy(t_ptr, m_data);
      delete []m_data;
      m_data = t_ptr;
   }

   return Removed;
}

/////////////////////////////////////////////////////////////////////////////
// Removes(int Location, int Total)
//
// Description: Remove string from m_data
//
// Location - index to the first character to remove.
// Total - total character to remove.
//
// TSTan   29 Jun 2002
/////////////////////////////////////////////////////////////////////////////
void CTString::Removes(int Location, int Total)
{
	while(Total--)
		RemoveAt(Location);
}

/////////////////////////////////////////////////////////////////////////////
// RemoveLeft(int Characters)
//
// Description: Remove string from left of string
//
// Characters - Total number of characters to be removed.
//
// TSTan   28 Oct 2000
/////////////////////////////////////////////////////////////////////////////
void CTString::RemoveLeft(int Characters)
{
   int Idx;
   for(Idx = 0; Idx < Characters; Idx++)
      RemoveAt(0);
}

/////////////////////////////////////////////////////////////////////////////
// RemoveWord()
//
// Description: Remove a word from left
//
// Return the removed word.
//
// TSTan   10 Mar 2002
/////////////////////////////////////////////////////////////////////////////
CTString &CTString::RemoveWord()
{
	TrimLeft();
	m_TempStorage = m_data;
	
	int space = GetChar(' ');
	int tab = GetChar('\t');
	if(tab != -1 && tab < space)
		space = tab;

	if(space > -1)
	{
		m_TempStorage.SetAt(space, 0);
		RemoveLeft(space + 1);
	}

	return m_TempStorage;
}

/////////////////////////////////////////////////////////////////////////////
// Replace(char original, char newchar)
//
// Description: Replace all 'original' to 'newchar' in the string.
//
// original - original character in the string.
// newchar - substituting character
//
// Return total characters replaced in the string.
//
// TSTan   10 Mar 2002
/////////////////////////////////////////////////////////////////////////////
int CTString::Replace(char original, char newchar)
{
   int idx, total;
	
	idx = total = 0;

	while(m_data[idx])
	{
		if(m_data[idx] == original)
		{
			m_data[idx] = newchar;
			total++;
		}
		idx++;
	}
	return total;
}

/////////////////////////////////////////////////////////////////////////////
// Replace(char *oldstr, char *newstr)
//
// Description: Replace all 'oldstr' to 'newstr' in the string.
//
// oldstr - original string in the string.
// newstr - substituting string
//
// Return total replacement occurrence in the string.
//
// TSTan   29 Jun 2002
/////////////////////////////////////////////////////////////////////////////
int CTString::Replace(char *oldstr, char *newstr)
{
	int newlength = strlen(newstr);
	int oldlength = strlen(oldstr);
	int idx = 0;
	int occur = 0;

	while(m_data[idx])
	{
		idx = Find(oldstr, idx);
		if(idx == -1)
			return occur;

		Removes(idx, oldlength);

		InsertAt(idx, newstr);
		idx += newlength;
		occur++;
	}

	return occur;
}

/////////////////////////////////////////////////////////////////////////////
// SetAt(int Location, char Character)
//
// Description: Set string at index Location to Character.
//
// Location - Offset from start of string to be replaced.
// Character - Character to replace to.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
void CTString::SetAt(int Location, char Character)
{
   if(Location < Size() && Location >= 0)
      m_data[Location] = Character;
}

/////////////////////////////////////////////////////////////////////////////
// SetFlag(int flag)
//
// Description: Set flag controlling string behaviour.
//
// flag - flag bit to set.
//
// TSTan   15 Sep 2002
/////////////////////////////////////////////////////////////////////////////
void CTString::SetFlag(int flag)
{
	m_flag |= flag;
}

/////////////////////////////////////////////////////////////////////////////
// Size()
//
// Description: Length of string without terminating NULL.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
int CTString::Size()
{
   return strlen(m_data);
}

/////////////////////////////////////////////////////////////////////////////
// ResetFlag(int flag)
//
// Description: Turn off behaviour controlling flag(s).
//
// flag - flag bit to turn off.
//
// TSTan   15 Sep 2002
/////////////////////////////////////////////////////////////////////////////
void CTString::ResetFlag(int flag)
{
	m_flag &= ~flag;
}

/////////////////////////////////////////////////////////////////////////////
// ToDouble
//
// Description: Convert the string indexed by Location to double.
//
// TSTan   11 Feb 2002
/////////////////////////////////////////////////////////////////////////////
double CTString::ToDouble(int Location)
{
	if(Location >= Size() || Location < 0)
		return 0.0;

	int Index, PosOrNeg, EPosOrNeg, Exponent;
	double Value;
	Value = 0.0;
	Exponent = 0;
	Index = Location;
	EPosOrNeg = PosOrNeg = 1;
	
	if(m_data[Index] == '-')
	{
		Index++;
		PosOrNeg = -1;
	}

	int ProcState = 0;	// 0 - Mantisa(>1), 1 - Mantisa (<1),
								// 2 - Exponent
	double Div = 0.1;
	do
	{
		if(m_data[Index] >= '0' && m_data[Index] <= '9')
		{
			switch(ProcState)
			{
			case 0:
				Value *= 10.0;
				Value += m_data[Index] - '0';
				break;
			case 1:
				Value += (m_data[Index] - '0') * Div;
				Div /= 10.0;
				break;
			case 2:
				Exponent *= 10;
				Exponent += m_data[Index] - '0';
				break;
			}
		}
		else if(m_data[Index] == '.' && !ProcState)
			ProcState = 1;
		else if((m_data[Index] == 'e' || m_data[Index] == 'E') &&
			ProcState != 2)
		{
			if(m_data[Index + 1] == '-')
			{
				EPosOrNeg = -1;
				Index++;
			}
			else if(m_data[Index + 1] == '+')
				Index++;

			ProcState = 2;
		}
		else
			break;

		Index++;

	}while(1);

	Value *= PosOrNeg;
	Value *= pow(10.0, EPosOrNeg * Exponent);

	return Value;
}	

/////////////////////////////////////////////////////////////////////////////
// ToInteger
//
// Description: Convert the string indexed by Location to integer
//
// TSTan   14 Mac 2001
/////////////////////////////////////////////////////////////////////////////
int CTString::ToInteger(int Location)
{
	if(Location >= Size() || Location < 0)
		return 0;

	int Value, Index, PosOrNeg;
	Value = 0;
	Index = Location;
	PosOrNeg = 1;
	
	if(m_data[Index] == '-')
	{
		Index++;
		PosOrNeg = -1;
	}

	do
	{
		if(m_data[Index] >= '0' && m_data[Index] <= '9')
		{
			Value *= 10;
			Value += m_data[Index] - '0';
		}
		else
			break;

		Index++;

	}while(1);

	Value *= PosOrNeg;

	return Value;
}	

/////////////////////////////////////////////////////////////////////////////
// ToHexInteger
//
// Description: Convert the string indexed by Location from Hex to integer
//
// TSTan   27 May 2001
/////////////////////////////////////////////////////////////////////////////
int CTString::ToHexInteger(int Location)
{
	if(Location >= Size() || Location < 0)
		return 0;

	int Value, Index;
	Value = 0;
	Index = Location;
	
	do
	{
		if(m_data[Index] >= '0' && m_data[Index] <= '9')
		{
			Value <<= 4;
			Value += m_data[Index] - '0';
		}
		else if(m_data[Index] >= 'a' && m_data[Index] <= 'f')
		{
			Value <<= 4;
			Value += m_data[Index] - 'a' + 10;
		}
		else if(m_data[Index] >= 'A' && m_data[Index] <= 'F')
		{
			Value <<= 4;
			Value += m_data[Index] - 'A' + 10;
		}
		else
			break;

		Index++;

	}while(1);

	return Value;
}	

/////////////////////////////////////////////////////////////////////////////
// ToUpper()
//
// Description: Change the whole string to upper case.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
void CTString::ToUpper()
{
   char *Ptr = m_data;
   while(*Ptr)
   {
      *Ptr = toupper(*Ptr);
      Ptr++;
   }
}

/////////////////////////////////////////////////////////////////////////////
// operator=(char *string)
//
// Description: Make this object containing string
//
// string - Pointer to string kept in char array.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
CTString &CTString::operator=(char *string)
{
   int Length = strlen(string) + 1;

   if(m_Current < Length ||   // not enough memory
      ((m_Current > m_Limit) && (m_Current > Length))) // wasted memory
   {
      delete []m_data;
      m_Current = Length;
      m_data = new char[m_Current];
   }

   strcpy(m_data, string);
   return *this;
}

/////////////////////////////////////////////////////////////////////////////
// operator=(CTString &string)
//
// Description: Make this object image of string.
//
// string - Reference to source string.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
CTString &CTString::operator=(CTString &string)
{
   m_Limit = string.m_Limit;
	m_precision = string.m_precision;
	m_width = string.m_width;
   *this = string.m_data;
   return *this;
}

/////////////////////////////////////////////////////////////////////////////
// operator==(char *string)
//
// Description: Check whether the string is the same with string pointed.
//
// string - Pointer to string kept in char array.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
bool CTString::operator==(char *string)
{
   return !strcmp(m_data, string);
}

/////////////////////////////////////////////////////////////////////////////
// operator==(CTString &string)
//
// Description: Check whether the string is the same with this object.
//
// string - Reference to string to be compared.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
bool CTString::operator==(CTString &string)
{
   return *this == string.m_data;
}

/////////////////////////////////////////////////////////////////////////////
// operator>(CTString &string)
//
// Description: True if this > string (alphabetically, length and lowercase
//              first).
//
// string - Reference to the second string.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
bool CTString::operator>(CTString &string)
{
   return strcmp(m_data, string.m_data) < 0;
}

/////////////////////////////////////////////////////////////////////////////
// operator<(CTString &string)
//
// Description: True if this < string (alphabetically, length and lowercase).
//
// string - Reference to the other string.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
bool CTString::operator<(CTString &string)
{
   return strcmp(m_data, string.m_data) > 0;
}

/////////////////////////////////////////////////////////////////////////////
// operator>=(CTString &string)
//
// Description: True if this >= string.
//
// string - Reference to the second string.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
bool CTString::operator>=(CTString &string)
{
   return (*this == string) || (*this > string);
}

/////////////////////////////////////////////////////////////////////////////
// operator<=(CTString &string)
//
// Description: True if this <= string.
//
// string - Reference to the second string.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
bool CTString::operator<=(CTString &string)
{
   return (*this == string) || (*this < string);
}

/////////////////////////////////////////////////////////////////////////////
// operator[](int index)
//
// Description: Return character offset from start of string.
//
// index - offset from start to get.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
char CTString::operator[](int index)
{
   return m_data[index];
}

/////////////////////////////////////////////////////////////////////////////
// operator+=(char character)
//
// Description: Append one character to the end of string.
//
// character - Character to be appended.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
CTString &CTString::operator+=(char character)
{
   int NewLength = strlen(m_data) + 2;
   char *t_ptr;
   if(m_Current < NewLength)
   {
      m_Current = NewLength;
      t_ptr = new char[m_Current];
      strcpy(t_ptr, m_data);
      delete []m_data;
      m_data = t_ptr;
   }
   else
      t_ptr = m_data;

   while(*t_ptr) t_ptr++;  // Search to the end
   *t_ptr = character;
   t_ptr++;
   *t_ptr = '\0';

   return *this;
}

/////////////////////////////////////////////////////////////////////////////
// operator+=(char *string)
//
// Description: Concatenate string.
//
// string - String to be concatenated.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
CTString &CTString::operator+=(char *string)
{
   int NewLength = strlen(m_data) + strlen(string) + 1;

   char *t_ptr;
   if(m_Current < NewLength)
   {
      m_Current = NewLength;
      t_ptr = new char[m_Current];
      strcpy(t_ptr, m_data);
      delete []m_data;
      m_data = t_ptr;
   }
   else
      t_ptr = m_data;

   strcat(t_ptr, string);

   return *this;
}

/////////////////////////////////////////////////////////////////////////////
// operator+=(CTString &string)
//
// Description: Append string.
//
// string - String to be appended.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
CTString &CTString::operator+=(CTString &string)
{
   *this += string.m_data;
   return *this;
}

/////////////////////////////////////////////////////////////////////////////
// operator+=(int number)
//
// Description: Append integer(after converting to string) to the string.
//
// number - integer to be appended.
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
CTString &CTString::operator+=(int number)
{
   char Buffer[80];
   ostrstream OutBuffer(Buffer, 80);

   OutBuffer << setw(m_width) << number << '\0';
   
	*this += OutBuffer.str();
   return *this;
}

/////////////////////////////////////////////////////////////////////////////
// operator+=(float number)
//
// Description: Append float(after converting to string) to the string.
//
// number - integer to be appended.
//
// TSTan   29 Oct 2000
/////////////////////////////////////////////////////////////////////////////
CTString &CTString::operator+=(float number)
{
   char Buffer[80];
   ostrstream OutBuffer(Buffer, 80);

	bool maxfault = false;
	int minusdec;
	minusdec = m_precision > 0? m_precision + 1: 0;
	if(m_width - minusdec > 0)
		maxfault = (number >= pow(10, minusdec))? true: false;

	if((m_flag & str_always_sci) ||	// Always use scientific
		(((number < pow(10, -m_precision) && number != 0.0)
		|| maxfault)	// Use scientific if
		&& (m_flag & str_scientific)))						// width constraint
		OutBuffer << setiosflags(ios::showpoint | ios::scientific)
		   << setprecision(m_precision) << setw(m_width) << number << '\0';
	else
	   OutBuffer << setiosflags(ios::showpoint | ios::fixed)
		   << setprecision(m_precision) << setw(m_width) << number << '\0';

   *this += OutBuffer.str();
   return *this;
}

/////////////////////////////////////////////////////////////////////////////
// operator+=(double number)
//
// Description: Append float(after converting to string) to the string.
//
// number - integer to be appended.
//
// TSTan   13 Feb 2002
/////////////////////////////////////////////////////////////////////////////
CTString &CTString::operator+=(double number)
{
   char Buffer[80];
   ostrstream OutBuffer(Buffer, 80);

	bool maxfault = false;
	int minusdec;
	minusdec = m_precision > 0? m_precision + 1: 0;
	if(m_width - minusdec > 0)
		maxfault = (number >= pow(10, minusdec))? true: false;

	if((m_flag & str_always_sci) ||	// Always use scientific
		(((number < pow(10, -m_precision) && number != 0.0)
		|| maxfault)	// Use scientific if
		&& (m_flag & str_scientific)))						// width constraint
		OutBuffer << setiosflags(ios::showpoint | ios::scientific)
		   << setprecision(m_precision) << setw(m_width) << number << '\0';
	else
	   OutBuffer << setiosflags(ios::showpoint | ios::fixed)
		   << setprecision(m_precision) << setw(m_width) << number << '\0';

   *this += OutBuffer.str();
   return *this;
}

/////////////////////////////////////////////////////////////////////////////
// StripLeftChars(int Characters)
//
// Description: Extract number of Characters from left and return.
//					 Chars in the original string removed.
//
// TSTan   11 Feb 2002
/////////////////////////////////////////////////////////////////////////////
CTString &CTString::StripLeftChars(int Characters)
{
	m_TempStorage = *this;

	m_TempStorage.TrimRight(Size() - Characters);
	TrimLeft(Characters);

	return m_TempStorage;
}

/////////////////////////////////////////////////////////////////////////////
// TrimLeft(int Characters)
//
// Description: Delete Characters from beginning of the string.
//
// Characters - <= 0: Delete blank spaces(' ' and '\t')
//					 Others: Number of characters to remove.
//
// TSTan   02 Jun 2001
/////////////////////////////////////////////////////////////////////////////
int CTString::TrimLeft(int Characters)
{
	if(Characters <= 0)
	{
		int idx = 0;
		while(((*this)[idx] == ' ' || (*this)[idx] == '\t') && (*this)[idx])
			idx++;

		strcpy(m_data, m_data + idx);
	}
	else if(Size() >= Characters)
		strcpy(m_data, m_data + Characters);
	else
		return 0;

	return 1;
}

/////////////////////////////////////////////////////////////////////////////
// TrimRight(int Characters)
//
// Description: Delete Characters from end of the string.
//
// Characters - <= 0: Delete blank spaces(' ' and '\t')
//					 Others: Number of characters to remove.
//
// TSTan   02 Jun 2001
/////////////////////////////////////////////////////////////////////////////
int CTString::TrimRight(int Characters)
{
	if(Characters <= 0)
	{
		int idx = Size() - 1;
		while(idx >= 0)
		{
			if((*this)[idx] == ' ' || (*this)[idx] == '\t')
			{
				idx--;
				continue;
			}
			break;
		}

		SetAt(idx + 1, 0);
	}
	else if(Size() >= Characters)
		SetAt(Size() - Characters, 0);
	else
		return 0;

	return 1;
}
	
/////////////////////////////////////////////////////////////////////////////
// operator<<(ostream &os, CTString &string)
//
// Description: Support stream insertion operation.
//
// os - stream to be inserted.
// string - CTString to insert
//
// TSTan   11 Oct 2000
/////////////////////////////////////////////////////////////////////////////
ostream &operator<<(ostream &os, CTString &string)
{
   os << string.m_data;

   return os;
}

CTString &operator<<(int &Value, CTString &String)
{
	while(String[0] && (String[0] < '0' || String[0] > '9'))
		String.RemoveAt(0);

	Value = String.ToInteger(0);

	while(String[0] && String[0] >= '0' && String[0] <= '9')
		String.RemoveAt(0);

	return String;
}